# -*- coding: utf-8 -*-
import unittest
import os
import numpy as npy

from skrf.media import CPW
from skrf.network import Network
from skrf.frequency import Frequency
import skrf as rf
from numpy.testing import assert_array_almost_equal, assert_allclose, run_module_suite


class CPWTestCase(unittest.TestCase):
    def setUp(self):
        self.files_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'qucs_prj'
            )        
        fname = os.path.join(self.files_dir, 'cpw.s2p')
        self.qucs_ntwk = rf.Network(fname)

        # create various examples
        self.freq = rf.Frequency(start=1, stop=20, npoints=21, unit='GHz')
        # infinite dielectric substrate, infinitly thin metal
        self.cpw1 = CPW(frequency=self.freq, w=40e-6, s=20e-6, ep_r=3)
        # infinite GaAs substrate, infinitly thin metal
        self.cpw2 = CPW(frequency=self.freq, w=75e-6, s=50e-6, ep_r=12.9)
        # infinite GaAs substrate, finite metal thickess
        # TODO: not used yet
        self.cpw3 = CPW(frequency=self.freq, w=75e-6, s=50e-6, ep_r=12.9, t=1e-6)

    def test_qucs_network(self):
        """
        Test against the Qucs project results
        TODO : finalize
        """
        # create an equivalent skrf network
        cpw = CPW(frequency=self.freq, w=75e-6, s=50e-6, ep_r=12.9, rho=0.22e-6)
        ntw = self.cpw2.thru(z0=50)**cpw.line(d=1, unit='m')**self.cpw2.thru(z0=50)
        # self.qucs_ntwk.plot_s_db()
        # ntw.plot_s_db()

    def test_Z0(self):
        """
        Test the CPW Characteristic Impedances
        """
        assert_array_almost_equal(self.cpw1.Z0, 85.25, decimal=3)
        
    def test_eps_eff(self):
        """
        Test the effective permittivity of CPW
        """
        assert_array_almost_equal(self.cpw1.ep_re, 2.00, decimal=3)
        assert_array_almost_equal(self.cpw2.ep_re, 6.95, decimal=3)        
        
    def test_Z0_vs_f(self):
        """
        Test the CPW Characteristic Impedance vs frequency. 
        
        Reference data comes from Qucs Documentation (Fig 12.2)
        """        
        w_over_s_qucs, Z0_qucs = npy.loadtxt(
            os.path.join(self.files_dir, 'cpw_qucs_ep_r9dot5.csv'), 
            delimiter=';', unpack=True)
               
        w = 1
        Z0 = []
        for w_o_s in w_over_s_qucs:
            _cpw = CPW(frequency=self.freq[0], w=w, s=w/w_o_s, ep_r=9.5)
            Z0.append(_cpw.Z0[0].real)
            
        # all to a 3% relative difference
        # this is quite a large discrepency, but I extracted the ref values from the plot
        # one could do better eventually by extracting values from Qucs directly
        rel_diff = (Z0_qucs-npy.array(Z0))/Z0_qucs
        assert_allclose(rel_diff  - 3/100, 0, atol=0.1)

if __name__ == "__main__":
    # Launch all tests
    run_module_suite()


import warnings
from collections import OrderedDict, Iterable

import numpy as np
import skrf
import pyvisa

from . import abcvna
from . import keysight_pna_scpi


class PNA(abcvna.VNA):
    """
    Class for modern Keysight/Agilent Peformance Network Analyzers
    """

    DEFAULT_VISA_ADDRESS = "GPIB::16::INSTR"
    NAME = "Keysight PNA"
    NPORTS = 2
    NCHANNELS = 32
    SCPI_VERSION_TESTED = 'A.09.20.08'

    def __init__(self, address=DEFAULT_VISA_ADDRESS, **kwargs):
        """
        initialization of PNA Class

        Parameters
        ----------
        address : str
            visa resource string (full string or ip address)
        kwargs : dict
            interface (str), port (int), timeout (int),
        :param address:
        :param kwargs:
        """
        super(PNA, self).__init__(address, **kwargs)
        self.resource.timeout = kwargs.get("timeout", 2000)
        self.scpi = keysight_pna_scpi.SCPI(self.resource)
        # self.use_binary()
        self.use_ascii()

    def use_binary(self):
        """setup the analyzer to transfer in binary which is faster, especially for large datasets"""
        self.resource.write(':FORM:BORD SWAP')
        self.resource.write(':FORM:DATA REAL,64')
        self.resource.values_format.use_binary(datatype='d', is_big_endian=False, container=np.array)

    def use_ascii(self):
        self.resource.write(':FORM:DATA ASCII')
        self.resource.values_format.use_ascii(converter='f', separator=',', container=np.array)

    @property
    def echo(self):
        return self.scpi.echo

    @echo.setter
    def echo(self, onoff):
        if onoff in (1, True):
            self.scpi.echo = True
        elif onoff in (0, False):
            self.scpi.echo = False
        else:
            raise warnings.warn("echo must be a boolean")

    @property
    def active_channel(self):
        old_timeout = self.resource.timeout
        self.resource.timeout = 500
        try:
            channel = self.scpi.query_active_channel()
        except pyvisa.VisaIOError:
            print("No channel active, using 1")
            channel = 1
        finally:
            self.resource.timeout = old_timeout
        return channel

    @active_channel.setter
    def active_channel(self, channel):
        """
        Set the active channel on the analyzer

        Parameters
        ----------
        channel : int

        Notes
        -----
        There is no specific command to activate a channel, so we ask which channel we want and then activate the first
        trace on that channel.  We do this because if the analyzer gets into a state where it doesn't recognize
        any activated measurements, the get_snp_network method will fail, and possibly others as well.  That is why in
        some methods you will see the fillowing line:
        self.active_channel = channel = kwargs.get("channel", self.active_channel)
        this way we force this property to be set, even if it just resets itself to the same value, but then a trace
        will become active and our get_snp_network method will succeed.
        """
        # TODO: Good chance this will fail if no measurement is on the set channel, need to think about that...
        mnum = self.scpi.query_meas_number_list(channel)[0]
        self.scpi.set_selected_meas_by_number(channel, mnum)
        return

    def sweep(self, **kwargs):
        """
        Initialize a fresh sweep of data on the specified channels

        Parameters
        ----------
        kwargs : dict
            channel ("all", int or list of channels), timeout (milliseconds)

        trigger a fresh sweep on the specified channels (default is "all" if no channel specified)
        Autoset timeout and sweep mode based upon the analyzers current averaging setting,
        and then return to the prior state of continuous trigger or hold.
        """
        self.resource.clear()
        self.scpi.set_trigger_source("IMM")
        original_timeout = self.resource.timeout

        # expecting either an int or a list of ints for the channel(s)
        channels_to_sweep = kwargs.get("channels", None)
        if not channels_to_sweep:
            channels_to_sweep = kwargs.get("channel", "all")
        if not type(channels_to_sweep) in (list, tuple):
            channels_to_sweep = [channels_to_sweep]
        channels = self.scpi.query_available_channels()

        for i, channel in enumerate(channels):
            sweep_mode = self.scpi.query_sweep_mode(channel)
            was_continuous = "CONT" in sweep_mode.upper()
            sweep_time = self.scpi.query_sweep_time(channel)
            averaging_on = self.scpi.query_averaging_state(channel)
            averaging_mode = self.scpi.query_averaging_mode(channel)

            if averaging_on and "SWE" in averaging_mode.upper():
                sweep_mode = "GROUPS"
                number_of_sweeps = self.scpi.query_averaging_count(channel)
                self.scpi.set_groups_count(channel, number_of_sweeps)
                number_of_sweeps *= self.NPORTS
            else:
                sweep_mode = "SINGLE"
                number_of_sweeps = self.NPORTS
            channels[i] = {
                "cnum": channel,
                "sweep_time": sweep_time,
                "number_of_sweeps": number_of_sweeps,
                "sweep_mode": sweep_mode,
                "was_continuous": was_continuous
            }
            self.scpi.set_sweep_mode(channel, "HOLD")
        timeout = kwargs.get("timeout", None)  # recommend not setting this variable, as autosetting is preferred

        try:
            for channel in channels:
                import time
                if "all" not in channels_to_sweep and channel["cnum"] not in channels_to_sweep:
                    continue  # default for sweep is all, else if we specify, then sweep
                if not timeout:  # autoset timeout based on sweep time
                    sweep_time = channel["sweep_time"] * channel[
                        "number_of_sweeps"] * 1000  # convert to milliseconds, and double for buffer
                    self.resource.timeout = max(sweep_time * 2, 5000)  # give ourselves a minimum 5 seconds for a sweep
                else:
                    self.resource.timeout = timeout
                self.scpi.set_sweep_mode(channel["cnum"], channel["sweep_mode"])
                self.wait_until_finished()
        finally:
            self.resource.clear()
            for channel in channels:
                if channel["was_continuous"]:
                    self.scpi.set_sweep_mode(channel["cnum"], "CONT")
            self.resource.timeout = original_timeout
        return

    def upload_twoport_calibration(self, cal, port1=1, port2=2, **kwargs):
        """
        upload a calibration to the vna, and set correction on all measurements
        
        Parameters
        ----------
        cal : skrf.Calibration
        port1: int
        port2: int

        calibration error terms reference
            # forward = (1, 1), reverse = (2, 2)
            "directivity": "EDIR",
            "source match": "ESRM",
            "reflection tracking": "ERFT",
            
            # forward = (2, 1), reverse = (1, 2)
            "load match": "ELDM",
            "transmission tracking": "ETRT"
            "isolation": "EXTLK"
        
        """
        self.active_channel = channel = kwargs.get("channel", self.active_channel)

        calname = kwargs.get("calname", "skrf_12term_cal")
        calibrations = self.scpi.query_calset_catalog(cnum=channel, form="NAME")
        if calname in calibrations:
            self.scpi.set_delete_calset(cnum=channel, calset_name=calname)
        self.scpi.set_create_calset(cnum=channel, calset_name=calname)

        cfs = dict()
        for coef, data in cal.coefs_12term.items():
            cfs[coef] = skrf.mf.complex2Scalar(data)

        for eterm, coef in zip(("EDIR", "ESRM", "ERFT"), ("directivity", "source match", "reflection tracking")):
            self.scpi.set_calset_data(channel, eterm, port1, port1, eterm_data=cfs["forward " + coef])
            self.scpi.set_calset_data(channel, eterm, port2, port2, eterm_data=cfs["reverse " + coef])
        for eterm, coef in zip(("ELDM", "ETRT", "EXTLK"), ("load match", "transmission tracking", "isolation")):
            self.scpi.set_calset_data(channel, eterm, port2, port1, eterm_data=cfs["forward " + coef])
            self.scpi.set_calset_data(channel, eterm, port1, port2, eterm_data=cfs["reverse " + coef])

        self.scpi.set_active_calset(1, calname, True)

    def get_snp_network(self, ports, **kwargs):
        """
        return n-port network as an Network object

        Parameters
        ----------
        ports : Iterable
            a iterable of integers designating the ports to query
        kwargs : dict
            channel(int) [ default 'self.active_channel' ]
            sweep(bool) [default True]
            name(str) [defaut \"\"]
            f_unit(str) [ default \"GHz\" ]
            raw_data(bool)  [default False]

        Returns
        -------
        Network

        general function to take in a list of ports and return the full snp network as a Network object
        """
        self.resource.clear()

        # force activate channel to avoid possible errors:
        self.active_channel = channel = kwargs.get("channel", self.active_channel)

        sweep = kwargs.get("sweep", True)
        name = kwargs.get("name", "")
        f_unit = kwargs.get("f_unit", "GHz")
        raw_data = kwargs.get("raw_data", False)

        ports = [int(port) for port in ports] if type(ports) in (list, tuple) else [int(ports)]
        if not name:
            name = "{:}Port Network".format(len(ports))
        if sweep:
            self.sweep(channel=channel)

        npoints = self.scpi.query_sweep_n_points(channel)

        snp_fmt = self.scpi.query_snp_format()
        self.scpi.set_snp_format("RI")
        if raw_data is True:
            if self.scpi.query_channel_correction_state(channel):
                self.scpi.set_channel_correction_state(channel, False)
                data = self.scpi.query_snp_data(channel, ports)
                self.scpi.set_channel_correction_state(channel, True)
            else:
                data = self.scpi.query_snp_data(channel, ports)
        else:
            data = self.scpi.query_snp_data(channel, ports)
        self.scpi.set_snp_format(snp_fmt)  # restore the value before we got the RI data
        self.scpi.set_snp_format(snp_fmt)  # restore the value before we got the RI data

        nrows = int(len(data) / npoints)
        nports = int(np.sqrt((nrows - 1)/2))
        data = data.reshape([nrows, -1])

        fdata = data[0]
        sdata = data[1:]
        ntwk = skrf.Network()
        ntwk.frequency = skrf.Frequency.from_f(fdata, unit="Hz")
        ntwk.s = np.empty(shape=(sdata.shape[1], nports, nports), dtype=complex)
        for n in range(nports):
            for m in range(nports):
                i = n * nports + m
                ntwk.s[:, m, n] = sdata[i * 2] + 1j * sdata[i * 2 + 1]
        ntwk.frequency.unit = f_unit
        ntwk.name = name
        return ntwk

    def get_list_of_traces(self, **kwargs):
        self.resource.clear()
        traces = []
        channels = self.scpi.query_available_channels()
        for channel in channels:
            meas_list = self.scpi.query_meas_name_list(channel)
            if len(meas_list) == 1:
                continue  # if there isnt a single comma, then there aren't any measurments
            parameters = dict([(meas_list[k], meas_list[k + 1]) for k in range(0, len(meas_list) - 1, 2)])

            meas_numbers = self.scpi.query_meas_number_list()
            for mnum in meas_numbers:
                name = self.scpi.query_meas_name_from_number(mnum)
                item = {"name": name, "channel": channel, "measurement number": mnum,
                        "parameter": parameters.get(name, name)}
                item["label"] = "{:s} - Chan{:},Meas{:}".format(
                    item["parameter"], item["channel"], item["measurement number"])
                traces.append(item)
        return traces

    def get_traces(self, traces, **kwargs):
        """
        retrieve traces as 1-port networks from a list returned by get_list_of_traces

        Parameters
        ----------
        traces : list
            list of type that is exported by self.get_list_of_traces
        kwargs : dict
            sweep (bool), name_prefix (str)

        Returns
        -------
        list
            a list of 1-port networks representing the desired traces

        Notes
        -----
        There is no current way to distinguish between traces and 1-port networks within skrf
        """
        self.resource.clear()
        sweep = kwargs.get("sweep", False)

        name_prefix = kwargs.get("name_prefix", "")
        if name_prefix:
            name_prefix += " - "

        channels = OrderedDict()
        for trace in traces:
            ch = trace["channel"]
            if ch not in channels.keys():
                channels[ch] = {
                    "frequency": None,
                    "traces": list()}
            channels[ch]["traces"].append(trace)

        if sweep is True:
            self.sweep(channels=list(channels.keys()))

        traces = []
        for ch, ch_data in channels.items():
            frequency = ch_data["frequency"] = self.get_frequency()
            for trace in ch_data["traces"]:
                self.scpi.set_selected_meas_by_number(trace["channel"], trace["measurement number"])
                sdata = self.scpi.query_data(trace["channel"], "SDATA")
                s = sdata[::2] + 1j * sdata[1::2]
                ntwk = skrf.Network()
                ntwk.s = s
                ntwk.frequency = frequency
                ntwk.name = name_prefix + trace.get("parameter", "trace")
                traces.append(ntwk)
        return traces

    def get_frequency(self, **kwargs):
        """
        get an skrf.Frequency object for the current channel

        Parameters
        ----------
        kwargs : dict
            channel (int), f_unit (str)

        Returns
        -------
        skrf.Frequency
        """
        self.resource.clear()
        channel = kwargs.get("channel", self.active_channel)
        sweep_type = self.scpi.query_sweep_type(channel)
        if sweep_type in ["LIN", "LOG", "SEGM"]:
            freqs = self.scpi.query_sweep_data(channel)
        else:
            freqs = np.array([self.scpi.query_f_start(channel)])

        frequency = skrf.Frequency.from_f(freqs, unit="Hz")
        frequency.unit = kwargs.get("f_unit", "Hz")
        return frequency

    def set_frequency_sweep(self, f_start, f_stop, f_npoints, **kwargs):
        f_unit = kwargs.get("f_unit", "hz").lower()
        if f_unit != "hz":
            f_start = self.to_hz(f_start, f_unit)
            f_stop = self.to_hz(f_stop, f_unit)
        channel = kwargs.get("channel", self.active_channel)
        self.scpi.set_f_start(channel, f_start)
        self.scpi.set_f_stop(channel, f_stop)
        self.scpi.set_sweep_n_points(f_npoints)

    def get_switch_terms(self, ports=(1, 2), **kwargs):
        self.resource.clear()
        p1, p2 = ports

        self.active_channel = channel = kwargs.get("channel", self.active_channel)

        measurements = self.get_meas_list()
        max_trace = len(measurements)
        for meas in measurements:  # type: str
            try:
                trace_num = int(meas[0][-2:].replace("_", ""))
                if trace_num > max_trace:
                    max_trace = trace_num
            except ValueError:
                pass

        forward_name = "CH{:}_FS_P{:d}_{:d}".format(channel, p1, max_trace + 1)
        reverse_name = "CH{:}_RS_P{:d}_{:d}".format(channel, p2, max_trace + 2)

        self.create_meas(forward_name, 'a{:}b{:},{:}'.format(p2, p2, p1))
        self.create_meas(reverse_name, 'a{:}b{:},{:}'.format(p1, p1, p2))

        self.sweep(channel=channel)

        forward = self.get_measurement(mname=forward_name, sweep=False)  # type: skrf.Network
        forward.name = "forward switch terms"
        reverse = self.get_measurement(mname=reverse_name, sweep=False)  # type: skrf.Network
        reverse.name = "reverse switch terms"

        self.scpi.set_delete_meas(channel, forward_name)
        self.scpi.set_delete_meas(channel, reverse_name)
        return forward, reverse

    def get_measurement(self, mname=None, mnum=None, **kwargs):
        """
        get a measurement trace from the analyzer, specified by either name or number

        Parameters
        ----------
        mname : str
            the name of the measurement, e.g. 'CH1_S11_1'
        mnum : int
            the number of number of the measurement
        kwargs : dict
            channel (int), sweep (bool)

        Returns
        -------
        skrf.Network
        """
        if mname is None and mnum is None:
            raise ValueError("must provide either a measurement mname or a mnum")

        channel = kwargs.get("channel", self.active_channel)

        if type(mname) is str:
            self.scpi.set_selected_meas(channel, mname)
        else:
            self.scpi.set_selected_meas_by_number(channel, mnum)
        return self.get_active_trace_as_network(**kwargs)

    def get_active_trace_as_network(self, **kwargs):
        """
        get the active trace as a network object

        Parameters
        ----------
        kwargs : dict
            channel (int), sweep (bool)

        Returns
        -------
        skrf.Network
        """
        channel = self.active_channel
        sweep = kwargs.get("sweep", False)
        if sweep:
            self.sweep(channel=channel)

        ntwk = skrf.Network()
        sdata = self.scpi.query_data(channel)
        ntwk.s = sdata[::2] + 1j * sdata[1::2]
        ntwk.frequency = self.get_frequency(channel=channel)
        return ntwk

    def create_meas(self, mname, param, **kwargs):
        """
        Create a new measurement trace on the analyzer

        Parameters
        ----------
        mname: str
            name of the measurement  **WARNING**, not all names behave well
        param: str
            analyzer parameter, e.g.: S11 ; a1/b1,1 ; A/R1,1
        kwargs : dict
            channel(int)
        """
        channel = kwargs.get("channel", self.active_channel)
        self.scpi.set_create_meas(channel, mname, param)
        self.display_trace(mname)

    def display_trace(self, mname, **kwargs):
        """
        Display measurement name on the analyzer display window

        Parameters:
        mname : str
            the name of the measurement, e.g. 'CH1_S11_1'
        kwargs : dict
            channel(int), window_n(int), trace_n(int), display_format(str)

        Keyword Arguments
        ----------------
        display_format : str
            must be one of: MLINear, MLOGarithmic, PHASe, UPHase, IMAGinary, REAL, POLar, SMITh,
                            SADMittance, SWR, GDELay, KELVin, FAHRenheit, CELSius
        """
        channel = kwargs.get('channel', self.active_channel)
        window_n = kwargs.get("window_n", '')
        trace_n = kwargs.get("trace_n",
                             max(self.scpi.query_window_trace_numbers(window_n)) + 1)
        display_format = kwargs.get('display_format', 'MLOG')

        self.scpi.set_display_trace(window_n, trace_n, mname)

        self.scpi.set_selected_meas(channel, mname)
        self.scpi.set_display_format(channel, display_format)

    def get_meas_list(self, **kwargs):
        """
        Convenience function to return a nicely arranged list of the measurement, parameter catalogue

        Parameters
        ----------
        kwargs : dict
            channel : int

        Returns
        -------
        list
            list of tuples of the form: (name, measurement)

        Return a list of measurement names on all channels.
        If channel is provided to kwargs, then only measurements for that channel are queried
        """
        channel = kwargs.get("channel", self.active_channel)
        meas_list = self.scpi.query_meas_name_list(channel)
        if len(meas_list) == 1:
            return None  # if there isnt a single comma, then there arent any measurments
        return [(meas_list[k], meas_list[k + 1]) for k in range(0, len(meas_list) - 1, 2)]

    @property
    def ntraces(self):
        """
        Get the number of traces on the active channel

        Returns
        -------
        int
            The number of measurement traces that exist on the current channel

        Notes
        -----
        Note that this may not be the same as the number of traces displayed because a measurement may exist,
        but not be associated with a trace.
        """
        meas_list = self.scpi.query_meas_number_list(self.active_channel)
        return 0 if meas_list is None else len(meas_list)


class PNAX(PNA):
    NAME = "Keysight PNA-X"
    NPORTS = 4
    NCHANNELS = 32



'''
skrf is an object-oriented approach to microwave engineering,
implemented in Python. 
'''
# Python 3 compatibility
from __future__ import absolute_import, print_function, division
from six.moves import xrange 

__version__ = '0.15.4'
## Import all  module names for coherent reference of name-space
#import io


from . import frequency
from . import network
from . import networkSet
from . import media
from . import circuit

from . import calibration
# from . import plotting
from . import mathFunctions
from . import tlineFunctions
from . import taper
from . import constants
from . import util
from . import io
from . import instances


# Import contents into current namespace for ease of calling
from .frequency import *
from .network import *
from .networkSet import *
from .calibration import *
from .util import *
from .circuit import *
# from .plotting import  *
from .mathFunctions import *
from .tlineFunctions import *
from .io import * 
from .constants import * 
from .taper import * 
from .instances import * 

# Try to import vi, but if except if pyvisa not installed
try:
    import vi
    from vi import *
except(ImportError):
    pass

# try to import data but if it fails whatever. it fails if some pickles 
# dont unpickle. but its not important
try:
    from . import data
except:
    pass 



## built-in imports
from copy import deepcopy as copy


## Shorthand Names
F = Frequency
N = Network
NS = NetworkSet
C = Circuit
lat = load_all_touchstones
# saf  = save_all_figs
saf = None
stylely = None


def setup_pylab():
    try:
        import matplotlib
    except ImportError:
        print("matplotlib not found while setting up plotting")
        return False

    from . import plotting
    plotting.setup_matplotlib_plotting()

    global saf, stylely
    saf = plotting.save_all_figs
    stylely = plotting.stylely
    return True


def setup_plotting():
    plotting_environment = os.environ.get('SKRF_PLOT_ENV', "pylab").lower()

    if plotting_environment == "pylab":
        setup_pylab()
    elif plotting_environment == "pylab-skrf-style":
        if setup_pylab():
            stylely()
    # elif some different plotting environment
        # set that up

setup_plotting()




import unittest
import os
import numpy as npy

import skrf as rf


class ConvenienceTestCase(unittest.TestCase):
    '''
    '''
    def setUp(self):
        '''
        '''
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/'
        self.hfss_oneport_file = os.path.join(self.test_dir, 'hfss_oneport.s1p')
        self.hfss_twoport_file = os.path.join(self.test_dir, 'hfss_twoport.s2p')
        self.hfss_threeport_file=os.path.join(self.test_dir, 'hfss_threeport_DB.s3p')
        self.hfss_threeport_file_50ohm=os.path.join(self.test_dir, 'hfss_threeport_DB_50Ohm.s3p')
        self.hfss_18dot2 = os.path.join(self.test_dir, 'hfss_18.2.s3p')
        self.hfss_8ports = os.path.join(self.test_dir, 'hfss_19.2.s8p')
        self.hfss_10ports = os.path.join(self.test_dir, 'hfss_19.2.s10p')
        self.ntwk1 = rf.Network(os.path.join(self.test_dir, 'ntwk1.s2p'))
        self.ntwk2 = rf.Network(os.path.join(self.test_dir, 'ntwk2.s2p'))
        self.ntwk3 = rf.Network(os.path.join(self.test_dir, 'ntwk3.s2p'))

    def test_hfss_high_port_number(self):
        '''
        Check dimensions s, gamma and z0 of Network from HFSS .sNp files 
        
        HFSS exports .sNp files with N > 4 with Gamma and Z0 as comments and
        written on multiple lines. The additional lines start with a '!' that 
        need to be escaped to avoid numerical conversion error. 
        '''
        _3ports = rf.Network(self.hfss_18dot2)
        self.assertTrue(_3ports.s.shape[1:] == (3,3))
        self.assertTrue(_3ports.z0.shape[1] == 3)
        self.assertTrue(_3ports.gamma.shape[1] == 3)
        
        _8ports = rf.Network(self.hfss_8ports)
        self.assertTrue(_8ports.s.shape[1:] == (8,8))
        self.assertTrue(_8ports.z0.shape[1] == 8)
        self.assertTrue(_8ports.gamma.shape[1] == 8)

        _10ports = rf.Network(self.hfss_10ports)
        self.assertTrue(_10ports.s.shape[1:] == (10,10))
        self.assertTrue(_10ports.z0.shape[1] == 10)
        self.assertTrue(_10ports.gamma.shape[1] == 10)
        

    def test_hfss_touchstone_2_media(self):
        '''
        currently, this just tests the execution ability. it would
        be better to simulate a uniform line, in hfss and then confirm
        that the hfss network is same as the one generated by the
        media object this function returns
        '''
        med = rf.hfss_touchstone_2_media(self.hfss_oneport_file)[0]
        med.line(1)
        med_p1,med_p2 = rf.hfss_touchstone_2_media(self.hfss_twoport_file)
        med_p1.line(1)
        med_p2.line(1)
        
    def test_hfss_touchstone_renormalization(self):
        '''
        Scattering matrices are given for a given impedance z0, 
        which is usually assumed to be 50 Ohm, unless otherwise stated.
        
        Touchstone files are not necessarly indicating such impedances, 
        especially if they vary with frequency.
        
        HFSS Touchstone file format supports port informations (as an option) for gamma and z0
        When HFSS files are read with Network() (or hfss_touchstone_2_network()),
        the port informations are taken into account.
        '''
        # Comparing the S-params of the same device expressed with same z0 
        nw_50 = rf.Network(self.hfss_threeport_file_50ohm)
        nw = rf.Network(self.hfss_threeport_file)
        nw.renormalize(z_new=50)       
        self.assertTrue(npy.all(npy.abs(nw.s - nw_50.s) < 1e-6))
        
    def test_is_hfss_touchstone(self):
        '''
        Test if Touchstone files have been generated by HFSS or not
        '''
        # Touchstone file generated by HFSS      
        self.assertTrue(rf.Touchstone(self.hfss_oneport_file).is_from_hfss() )
        self.assertTrue(rf.Touchstone(self.hfss_twoport_file).is_from_hfss() )
        self.assertTrue(rf.Touchstone(self.hfss_threeport_file).is_from_hfss() )
        self.assertTrue(rf.Touchstone(self.hfss_18dot2).is_from_hfss())
        self.assertTrue(rf.Touchstone(self.hfss_8ports).is_from_hfss())
        self.assertTrue(rf.Touchstone(self.hfss_10ports).is_from_hfss())
        # Touchstone file not from HFSS       
        self.assertFalse(rf.Touchstone(os.path.join(self.test_dir, 'ntwk1.s2p')).is_from_hfss() )
        
    def test_hfss_touchstone_2_network(self):
        '''
        Test the conversion into a Network of HFSS-generated touchstone files
        '''
        nw_hfss_wo_z0 = rf.Network(os.path.join(self.test_dir, 'hfss_threeport_MA_without_gamma_z0_50Ohm.s3p'))
        nw_hfss_50 = rf.Network(os.path.join(self.test_dir, 'hfss_threeport_MA_50Ohm.s3p'))
        nw_hfss_z0 = rf.Network(os.path.join(self.test_dir, 'hfss_threeport_MA.s3p'))
        # Test if the values read are the same 
        self.assertTrue(npy.allclose(nw_hfss_50.s, nw_hfss_wo_z0.s))
        nw_hfss_z0.renormalize(50)
        self.assertTrue(npy.allclose(nw_hfss_50.s, nw_hfss_z0.s))
        
    def test_cst_touchstone_2_network(self):
        '''
        Test the conversion into a Network of CST-generated touchstone file
        '''
        nw_cst_4ports = rf.Network(os.path.join(self.test_dir, 'cst_example_4ports.s4p'))
        nw_cst_6ports = rf.Network(os.path.join(self.test_dir, 'cst_example_6ports.s6p'))
        
    def test_cst_touchstone_V2_as_V1_2_network(self):
        '''
        Test the conversion into a Network of a CST-generated 
        touchstone V2 format file (.ts) saved like a touchstone V1 file (.sNp)
        '''
        nw_cst_6ports = rf.Network(os.path.join(self.test_dir, 'cst_example_6ports_V2.s6p'))
 
    def test_cst_touchstone_V2_2_network(self):
        '''
        Test the conversion into a Network of a CST-generated touchstone V2 format file (.ts) 
        '''
        nw_cst_6ports = rf.Network(os.path.join(self.test_dir, 'cst_example_6ports_V2.ts'))       
        
    def test_Agilent_touchstone_4ports(self):		
        '''		
        Try reading an Agilent touchstone 4-ports measurement file		
        '''		
        filename = 'Agilent_E5071B.s4p'		
        ntwk = rf.Network(os.path.join(self.test_dir, filename))				          
        # Check if port characteric impedance is correctly parsed        
        self.assertTrue(npy.isclose(npy.unique(ntwk.z0), 75)) 
        
        self.assertTrue(npy.allclose(ntwk.s_db[0][1], # check s2n_mag
                                    [-5.252684e+001, -2.278388e-001, -4.435702e+001, -8.235984e+001]))
        self.assertTrue(npy.allclose(ntwk.s_deg[0][1], # check s2n_deg
                                    [-1.350884e+002, 8.767636e+001,	-1.585657e+002,	7.708928e+001]))
        
    def test_RS_touchstone_4ports(self):		
        '''		
        Try reading an R&S touchstone 4-ports measurement file		
        '''		
        filename = 'RS_ZNB8.s4p'		
        ntwk = rf.Network(os.path.join(self.test_dir, filename))		
        # Check if port characteric impedance is correctly parsed		
        self.assertTrue(npy.isclose(npy.unique(ntwk.z0), 50))		
        # For this specific file, the port#1 min return loss is @55.5MHz		
        self.assertTrue(ntwk.frequency.f[npy.argmin(ntwk.s11.s_mag)], 55.5e6)

        self.assertTrue(npy.allclose(ntwk.s_re[0][2], # check s3n_re
                                    [-9.748145748042028E-6, 5.737806652221101E-6, -7.283138400961303E-1,  -7.202238521877286E-6]))
        self.assertTrue(npy.allclose(ntwk.s_im[0][2], # check s3n_im
                                    [4.457944078457155E-6, 5.341399484369366E-6, -4.531402467395991E-1, 5.667857998796495E-7]))                 
        

if __name__ == "__main__":
    from numpy.testing import run_module_suite
    # Launch all tests
    run_module_suite()            



# -*- coding: utf-8 -*-
import skrf as rf
import numpy as np
import unittest
import os, sys
from numpy.testing import assert_array_almost_equal, run_module_suite


class CircuitTestConstructor(unittest.TestCase):
    '''
    Various tests on the Circuit constructor.
    '''
    def setUp(self):
        # Importing network examples
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/'
        self.ntwk1 = rf.Network(os.path.join(self.test_dir, 'ntwk1.s2p'))
        self.ntwk2 = rf.Network(os.path.join(self.test_dir, 'ntwk2.s2p'))
        self.freq = self.ntwk1.frequency
        # circuit external ports
        self.port1 = rf.Circuit.Port(self.freq, name='Port1')
        self.port2 = rf.Circuit.Port(self.freq, name='Port2')

    def test_all_networks_have_name(self):
        '''
        Check that a Network without name raises an exception
        '''
        _ntwk1 = self.ntwk1.copy()
        connections = [[(self.port1, 0), (_ntwk1, 0)],
                       [(_ntwk1, 1), (self.ntwk2, 0)],
                       [(self.ntwk2, 1), (self.port2, 0)]]

        _ntwk1.name = []
        self.assertRaises(AttributeError, rf.Circuit, connections)

        _ntwk1.name = ''
        self.assertRaises(AttributeError, rf.Circuit, connections)

    def test_all_networks_have_same_frequency(self):
        '''
        Check that a Network with a different frequency than the other
        raises an exception
        '''
        _ntwk1 = self.ntwk1.copy()
        connections = [[(self.port1, 0), (_ntwk1, 0)],
                       [(_ntwk1, 1), (self.ntwk2, 0)],
                       [(self.ntwk2, 1), (self.port2, 0)]]

        _ntwk1.frequency = rf.Frequency(start=1, stop=1, npoints=1)
        self.assertRaises(AttributeError, rf.Circuit, connections)

    def test_s_active(self):
        '''
        Test the active s-parameter of a 2-ports network
        '''
        connections = [[(self.port1, 0), (self.ntwk1, 0)],
                       [(self.ntwk1, 1), (self.ntwk2, 0)],
                       [(self.ntwk2, 1), (self.port2, 0)]]
        circuit = rf.Circuit(connections)
        # s_act should be equal to s11 if a = [1,0]
        assert_array_almost_equal(circuit.s_active([1, 0])[:,0], circuit.s_external[:,0,0])
        # s_act should be equal to s22 if a = [0,1]
        assert_array_almost_equal(circuit.s_active([0, 1])[:,1], circuit.s_external[:,1,1])
                

class CircuitTestWilkinson(unittest.TestCase):
    '''
    Create a Wilkison power divider Circuit [#]_ and test the results
    against theoretical ones (obtained in [#]_)

    References
    ----------
    .. [#] https://en.wikipedia.org/wiki/Wilkinson_power_divider
    .. [#] P. Hallbjörner, Microw. Opt. Technol. Lett. 38, 99 (2003).


    '''
    def setUp(self):
        '''
        Circuit setup
        '''
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/'
        self.freq = rf.Frequency(start=1, stop=2, npoints=101)
        # characteristic impedance of the ports
        Z0_ports = 50

        # resistor
        self.R = 100
        self.line_resistor = rf.media.DefinedGammaZ0(frequency=self.freq, Z0=self.R)
        self.resistor = self.line_resistor.resistor(self.R, name='resistor')

        # branches
        Z0_branches = np.sqrt(2)*Z0_ports
        self.line_branches = rf.media.DefinedGammaZ0(frequency=self.freq, Z0=Z0_branches)
        self.branch1 = self.line_branches.line(90, unit='deg', name='branch1')
        self.branch2 = self.line_branches.line(90, unit='deg', name='branch2')

        # ports
        port1 = rf.Circuit.Port(self.freq, name='port1')
        port2 = rf.Circuit.Port(self.freq, name='port2')
        port3 = rf.Circuit.Port(self.freq, name='port3')

        # Connection setup
        self.connections = [
                   [(port1, 0), (self.branch1, 0), (self.branch2, 0)],
                   [(port2, 0), (self.branch1, 1), (self.resistor, 0)],
                   [(port3, 0), (self.branch2, 1), (self.resistor, 1)]
                ]

        self.C = rf.Circuit(self.connections)

        # theoretical results from ref P.Hallbjörner (2003)
        self.X1_nn = np.array([1 - np.sqrt(2), -1, -1])/(1 + np.sqrt(2))
        self.X2_nn = np.array([1 - np.sqrt(2), -3 + np.sqrt(2),
                               -1 - np.sqrt(2)])/(3 + np.sqrt(2))

        self.X1_m1 = 2 / (1 + np.sqrt(2))
        self.X1_m2 = np.sqrt(2) / (1 + np.sqrt(2))
        self.X1_m3 = np.sqrt(2) / (1 + np.sqrt(2))

        self.X2_m1 = 4 / (3 + np.sqrt(2))
        self.X2_m2 = 2*np.sqrt(2) / (3 + np.sqrt(2))
        self.X2_m3 = 2 / (3 + np.sqrt(2))

        self.X1 = np.array([[0, self.X1_m2, self.X1_m3],
                            [self.X1_m1, 0, self.X1_m3],
                            [self.X1_m1, self.X1_m2, 0]]) + np.diag(self.X1_nn)
        self.X2 = np.array([[0, self.X2_m2, self.X2_m3],
                            [self.X2_m1, 0, self.X2_m3],
                            [self.X2_m1, self.X2_m2, 0]]) + np.diag(self.X2_nn)

    def test_global_admittance(self):
        '''
        Check is Y is correct wrt to ref P.Hallbjörner (2003)
        '''
        Y1 = (1 + np.sqrt(2)) / 50
        Y2 = (3 + np.sqrt(2)) / 100

        assert_array_almost_equal(self.C._Y_k(self.connections[0]), Y1)
        assert_array_almost_equal(self.C._Y_k(self.connections[1]), Y2)
        assert_array_almost_equal(self.C._Y_k(self.connections[2]), Y2)

    def test_reflection_coefficients(self):
        '''
        Check if Xnn are correct wrt to ref P.Hallbjörner (2003)
        '''
        assert_array_almost_equal(self.C._Xnn_k(self.connections[0])[0], self.X1_nn)
        assert_array_almost_equal(self.C._Xnn_k(self.connections[1])[0], self.X2_nn)
        assert_array_almost_equal(self.C._Xnn_k(self.connections[2])[0], self.X2_nn)

    def test_transmission_coefficients(self):
        '''
        Check if Xmn are correct wrt to ref P.Hallbjörner (2003)
        '''
        assert_array_almost_equal(self.C._Xmn_k(self.connections[0])[0], np.r_[self.X1_m1, self.X1_m2, self.X1_m2])
        assert_array_almost_equal(self.C._Xmn_k(self.connections[1])[0], np.r_[self.X2_m1, self.X2_m2, self.X2_m3])
        assert_array_almost_equal(self.C._Xmn_k(self.connections[2])[0], np.r_[self.X2_m1, self.X2_m2, self.X2_m3])

    def test_sparam_individual_intersection_matrices(self):
        '''
        Testing the individual intersection scattering matrices X_k
        '''
        np.testing.assert_array_almost_equal(self.C._Xk(self.connections[0])[0], self.X1)
        np.testing.assert_array_almost_equal(self.C._Xk(self.connections[1])[0], self.X2)
        np.testing.assert_array_almost_equal(self.C._Xk(self.connections[2])[0], self.X2)

    def test_sparam_global_intersection_matrix(self):
        '''
        Testing the global intersection scattering matrix
        '''
        from scipy.linalg import block_diag
        assert_array_almost_equal(self.C.X[0], block_diag(self.X1, self.X2, self.X2) )

    def test_sparam_circuit(self):
        '''
        Testing the external scattering matrix
        '''
        S_theoretical = np.array([[0, 1, 1],
                                  [1, 0, 0],
                                  [1, 0, 0]]) * (-1j/np.sqrt(2))

        # extracting the external ports
        S_ext = self.C.s_external

        assert_array_almost_equal(S_ext[0], S_theoretical)

    def test_compare_with_skrf_wilkison(self):
        '''
        Create a Wilkinson power divider using skrf usual Network methods.
        '''
        z0_port = 50
        z0_lines = self.line_branches.z0[0]
        z0_R = self.line_resistor.z0[0]
        # require to create the three tees
        T0 = self.line_branches.splitter(3, z0=[z0_port, z0_lines, z0_lines])
        T1 = self.line_branches.splitter(3, z0=[z0_lines, z0_R, z0_port])
        T2 = self.line_branches.splitter(3, z0=[z0_lines, z0_R, z0_port])

        _wilkinson1 = rf.connect(T0, 1, self.branch1, 0)
        _wilkinson2 = rf.connect(_wilkinson1, 2, self.branch2, 0)
        _wilkinson3 = rf.connect(_wilkinson2, 0, T1, 0)
        _wilkinson4 = rf.connect(_wilkinson3, 0, T2, 0)
        _wilkinson5 = rf.connect(_wilkinson4, 1, self.resistor, 0)
        wilkinson = rf.innerconnect(_wilkinson5, 0, 3)

        ntw_C = self.C.network

        # the following is failing and I don't know why
        #assert_array_almost_equal(ntw_C.s_db, wilkinson.s_db)

        assert_array_almost_equal(ntw_C.z0, wilkinson.z0)

    def test_compare_with_designer_wilkinson(self):
        '''
        Compare the result with ANSYS Designer model

        Built as in https://www.microwaves101.com/encyclopedias/wilkinson-power-splitters
        '''
        designer_wilkinson = rf.Network(os.path.join(self.test_dir, 'designer_wilkinson_splitter.s3p'))
        ntw_C = self.C.network

        assert_array_almost_equal(ntw_C.s[0], designer_wilkinson.s[0], decimal=4)

    def test_s_active(self):
        '''
        Test the active s-parameter of a 3-ports network
        '''
        # s_act should be equal to s11 if a = [1,0,0]
        assert_array_almost_equal(self.C.network.s_active([1, 0, 0])[:,0], self.C.s_external[:,0,0])
        # s_act should be equal to s22 if a = [0,1,0]
        assert_array_almost_equal(self.C.network.s_active([0, 1, 0])[:,1], self.C.s_external[:,1,1])
        # s_act should be equal to s33 if a = [0,0,1]
        assert_array_almost_equal(self.C.network.s_active([0, 0, 1])[:,2], self.C.s_external[:,2,2])
        
class CircuitTestCascadeNetworks(unittest.TestCase):
    '''
    Build a circuit made of two Networks cascaded and compare the result
    to usual cascading of two networks.
    '''
    def setUp(self):
        # Importing network examples
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/'
        self.ntwk1 = rf.Network(os.path.join(self.test_dir, 'ntwk1.s2p'))
        self.ntwk2 = rf.Network(os.path.join(self.test_dir, 'ntwk2.s2p'))
        # ntwk3 is the cascade of ntwk1 and ntwk2, ie. self.ntwk1 ** self.ntwk2
        self.ntwk3 = rf.Network(os.path.join(self.test_dir, 'ntwk3.s2p'))
        self.freq = self.ntwk1.frequency
        # circuit external ports
        self.port1 = rf.Circuit.Port(self.freq, name='Port1')
        self.port2 = rf.Circuit.Port(self.freq, name='Port2')

    def test_cascade(self):
        '''
        Compare ntwk3 to the Circuit of ntwk1 and ntwk2.
        '''
        connections = [  [(self.port1, 0), (self.ntwk1, 0)],
                         [(self.ntwk1, 1), (self.ntwk2, 0)],
                         [(self.ntwk2, 1), (self.port2, 0)] ]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(circuit.s_external, self.ntwk3.s)

    def test_cascade2(self):
        '''
        Same thing with different ordering of the connections.
        Demonstrate that changing the connections setup order does not change
        the result.
        '''
        connections = [  [(self.port1, 0), (self.ntwk1, 0)],
                         [(self.ntwk2, 0), (self.ntwk1, 1)],
                         [(self.port2, 0), (self.ntwk2, 1)] ]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(circuit.s_external, self.ntwk3.s)

    def test_cascade3(self):
        '''
        Inverting the cascading network order
        Demonstrate that changing the connections setup order does not change
        the result (at the requirement that port impedance are the same).
        '''
        connections = [  [(self.port1, 0), (self.ntwk2, 0)],
                         [(self.ntwk2, 1), (self.ntwk1, 0)],
                         [(self.port2, 0), (self.ntwk1, 1)] ]
        circuit = rf.Circuit(connections)
        ntw = self.ntwk2 ** self.ntwk1
        assert_array_almost_equal(circuit.s_external, ntw.s)

class CircuitTestMultiPortCascadeNetworks(unittest.TestCase):
    '''
    Various 1-ports, 2-ports and 4-ports circuits and associated tests
    '''
    def test_1port_matched_load(self):
        '''
        Connect a matched load directly to the port
        '''
        freq = rf.Frequency(start=1, npoints=1)
        port1 = rf.Circuit.Port(freq,  name='port1')
        line = rf.media.DefinedGammaZ0(frequency=freq)
        match_load = line.match(name='match_load')
        
        cnx = [
            [(port1, 0), (match_load, 0)]
        ]
        cir = rf.Circuit(cnx)
        
        assert_array_almost_equal(match_load.s, cir.s_external)
    
    def test_1port_short(self):
        '''
        Connect a short directly to the port
        '''
        freq = rf.Frequency(start=1, npoints=1)
        port1 = rf.Circuit.Port(freq,  name='port1')
        line = rf.media.DefinedGammaZ0(frequency=freq)
        short = line.short(name='short')
        gnd1 = rf.Circuit.Ground(freq, name='gnd')
        # method 1 : use the Ground Network (which 2 port actually)
        cnx = [
            [(port1, 0), (gnd1, 0)]
        ]
        cir = rf.Circuit(cnx)
        assert_array_almost_equal(short.s, cir.s_external)
        # method 2 : use a short Network (1 port)
        cnx = [
            [(port1, 0), (short, 0)]
        ]
        cir = rf.Circuit(cnx)

        assert_array_almost_equal(short.s, cir.s_external)
        
    def test_1port_random_load(self):
        '''
        Connect a random load directly to the port
        '''
        freq = rf.Frequency(start=1, npoints=1)
        port1 = rf.Circuit.Port(freq,  name='port1')
        line = rf.media.DefinedGammaZ0(frequency=freq)
        gamma = np.random.rand(1,1) + 1j*np.random.rand(1,1)
        load = line.load(gamma, name='load')
        
        cnx = [
            [(port1, 0), (load, 0)]
        ]
        cir = rf.Circuit(cnx)

        assert_array_almost_equal(load.s, cir.s_external)
        
    def test_1port_matched_network_default_impedance(self):
        '''
        Connect a random 2 port network connected to a matched load
        '''
        freq = rf.Frequency(start=1, npoints=1)
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.rand(4).reshape(2, 2)
        line = rf.media.DefinedGammaZ0(frequency=freq)
        match_load = line.match(name='match_load')

        # classic connecting
        b = a ** match_load

        # Circuit connecting
        port1 = rf.Circuit.Port(freq,  name='port1')
        connections = [[(port1, 0), (a, 0)], [(a, 1), (match_load, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(b.s, circuit.s_external)

    def test_1port_matched_network_complex_impedance(self):
        '''
        Connect a 2 port network to a complex impedance.
        Both ports are complex.
        '''
        z01, z02 = 1-1j, 2+4j
        freq = rf.Frequency(start=1, npoints=1)
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.rand(4).reshape(2, 2)
        a.z0 = [z01, z02]
        line = rf.media.DefinedGammaZ0(frequency=freq, z0=z02)
        match_load = line.match(name='match_load')

        # classic connecting
        b = a ** match_load

        # Circuit connecting
        port1 = rf.Circuit.Port(freq, z0=z01, name='port1')
        connections = [[(port1, 0), (a,0)], [(a, 1), (match_load, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(b.s, circuit.s_external)

    def test_2ports_default_characteristic_impedance(self):
        '''
        Connect two 2-ports networks in a resulting  2-ports network,
        same default charact impedance (50 Ohm) for all ports
        '''
        freq = rf.Frequency(start=1, npoints=1)
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.rand(4).reshape(2, 2)

        b = rf.Network(name='b')
        b.frequency = freq
        b.s = np.random.rand(4).reshape(2, 2)

        # classic connecting
        c = rf.connect(a, 1, b, 0)

        # Circuit connecting
        port1 = rf.Circuit.Port(freq,  name='port1')
        port2 = rf.Circuit.Port(freq,  name='port2')

        connections = [[(port1, 0), (a, 0)],
                       [(a, 1), (b, 0)],
                       [(b, 1), (port2, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(c.s, circuit.s_external)

    def test_2ports_complex_characteristic_impedance(self):
        '''
        Connect two 2-ports networks in a resulting  2-ports network,
        same complex charact impedance (1+1j) for all ports
        '''
        z0 = 1 + 1j
        freq = rf.Frequency(start=1, npoints=1)
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.rand(4).reshape(2, 2)
        a.z0 = z0

        b = rf.Network(name='b')
        b.frequency = freq
        b.s = np.random.rand(4).reshape(2, 2)
        b.z0 = z0

        # classic connecting
        c = rf.connect(a, 1, b, 0)

        # Circuit connecting
        port1 = rf.Circuit.Port(freq, z0=z0, name='port1')
        port2 = rf.Circuit.Port(freq, z0=z0, name='port2')

        connections = [[(port1, 0), (a, 0)],
                       [(a, 1), (b, 0)],
                       [(b, 1), (port2, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(c.s, circuit.s_external)

    def test_2ports_different_characteristic_impedances(self):
        '''
        Connect two 2-ports networks in a resulting  2-ports network,
        different characteristic impedances for each network ports
        '''
        freq = rf.Frequency(start=1, npoints=1)
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.rand(4).reshape(2,2)
        a.z0 = [1, 2]  #  Z0 should never be zero

        b = rf.Network(name='b')
        b.frequency = freq
        b.s = np.random.rand(4).reshape(2,2)
        b.z0 = [11, 12]

        # classic connecting
        c = rf.connect(a, 1, b, 0)

        # Circuit connecting
        port1 = rf.Circuit.Port(freq, z0=1, name='port1')
        port2 = rf.Circuit.Port(freq, z0=12, name='port2')

        connections = [[(port1, 0), (a, 0)],
                       [(a, 1), (b, 0)],
                       [(b, 1), (port2, 0)] ]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(c.s, circuit.s_external)

    def test_4ports_default_characteristic_impedances(self):
        '''
        Connect two 4-ports networks in a resulting 4-ports network,
        with default characteric impedances
        '''
        freq = rf.Frequency(start=1, npoints=1)
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.rand(16).reshape(4, 4)

        b = rf.Network(name='b')
        b.frequency = freq
        b.s = np.random.rand(16).reshape(4, 4)

        # classic connecting
        c = rf.connect(a, 2, b, 0, 2)

        # Circuit connecting
        port1 = rf.Circuit.Port(freq, name='port1')
        port2 = rf.Circuit.Port(freq, name='port2')
        port3 = rf.Circuit.Port(freq, name='port3')
        port4 = rf.Circuit.Port(freq, name='port4')

        connections = [[(port1, 0), (a, 0)],
                       [(port2, 0), (a, 1)],
                       [(a, 2), (b, 0)],
                       [(a, 3), (b, 1)],
                       [(b, 2), (port3, 0)],
                       [(b, 3), (port4, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(c.s, circuit.s_external)

    def test_4ports_complex_characteristic_impedances(self):
        '''
        Connect two 4-ports networks in a resulting 4-ports network,
        with same complex characteric impedances
        '''
        z0 = 5 + 4j
        freq = rf.Frequency(start=1, npoints=1)
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.rand(16).reshape(4, 4)
        a.z0 = z0

        b = rf.Network(name='b')
        b.frequency = freq
        b.s = np.random.rand(16).reshape(4, 4)
        b.z0 = z0

        # classic connecting
        c = rf.connect(a, 2, b, 0, 2)

        # circuit connecting
        port1 = rf.Circuit.Port(freq, z0=z0, name='port1')
        port2 = rf.Circuit.Port(freq, z0=z0, name='port2')
        port3 = rf.Circuit.Port(freq, z0=z0, name='port3')
        port4 = rf.Circuit.Port(freq, z0=z0, name='port4')

        connections = [ [(port1, 0), (a, 0)],
                        [(port2, 0), (a, 1)],
                        [(a, 2), (b, 0)],
                        [(a, 3), (b, 1)],
                        [(b, 2), (port3, 0)],
                        [(b, 3), (port4, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(c.s, circuit.s_external)

    def test_4ports_different_characteristic_impedances(self):
        '''
        Connect two 4-ports networks in a resulting 4-ports network,
        with different characteristic impedances
        '''
        z0 = [1, 2, 3, 4]
        freq = rf.Frequency(start=1, npoints=1)
        a = rf.Network(name='a')
        a.frequency = freq
        a.s = np.random.rand(16).reshape(4, 4)
        a.z0 = z0

        b = rf.Network(name='b')
        b.frequency = freq
        b.s = np.random.rand(16).reshape(4, 4)
        b.z0 = [11, 12, 13, 14]

        # classic connecting
        _c = rf.connect(a, 2, b, 0)
        c = rf.innerconnect(_c, 2, 3)

        # Circuit connecting
        port1 = rf.Circuit.Port(freq, z0=1, name='port1')
        port2 = rf.Circuit.Port(freq, z0=2, name='port2')
        port3 = rf.Circuit.Port(freq, z0=13, name='port3')
        port4 = rf.Circuit.Port(freq, z0=14, name='port4')

        connections = [[(port1, 0), (a, 0)],
                       [(port2, 0), (a, 1)],
                       [(a, 2), (b, 0)],
                       [(a, 3), (b, 1)],
                       [(b, 2), (port3, 0)],
                       [(b, 3), (port4, 0)]]
        circuit = rf.Circuit(connections)

        assert_array_almost_equal(c.s, circuit.s_external)

    def test_shunt_element(self):
        '''
        Compare a shunt element network (here a capacitor) 
        '''
        freq = rf.Frequency(start=1, stop=2, npoints=101)
        line = rf.media.DefinedGammaZ0(frequency=freq, z0=50)
        # usual way
        cap_shunt_manual = line.shunt_capacitor(50e-12)
        
        # A Circuit way
        port1 = rf.Circuit.Port(frequency=freq, name='port1', z0=50)
        port2 = rf.Circuit.Port(frequency=freq, name='port2', z0=50)
        cap_shunt = line.capacitor(50e-12, name='cap_shunt')
        ground = rf.Circuit.Ground(frequency=freq, name='ground', z0=50)

        connections = [
            [(port1, 0), (cap_shunt, 0), (port2, 0)],
            [(cap_shunt, 1), (ground, 0)]
        ]
        
        # # Another possibility could have been without ground :
        # shunt_cap = line.shunt_capacitor(50e-12)
        # shunt_cap.name='shunt_cap'
        # connections = [
        #     [(port1, 0), (shunt_cap, 0)],
        #     [(shunt_cap ,1), (port2, 0)]
        # ]
        
        cap_shunt_from_circuit = rf.Circuit(connections).network

        assert_array_almost_equal(cap_shunt_manual.s, cap_shunt_from_circuit.s)

class CircuitTestVariableCoupler(unittest.TestCase):
    '''
    If we use 3 dB hybrid defined as :
                   ________
    Input     0 --|       |-- 1 Through
                  |       |
    Isolated  3 --|_______|-- 2 Coupled

    And if we connect two 3 dB hybrid with a phase shifter such as:
                 _________                                 _________
    Port#3 >- 0-|        |-1 -------------------------- 0-|        |-1 -< Port#2
                |  Hyb1  |                                |  Hyb2  |
    Port#0 >- 3-|________|-2 -- 0-[phase shifter]-1 -- 3-|________|-2 -< Port#1

    Then we have a variable coupler. The coupling factor can be adjusted
    by changing the phase of the phase shifter.

    The port order in this example is volontary complicated to make a good
    example.

    '''
    def setUp(self):
        self.freq = rf.Frequency(start=1.5, stop=1.5, npoints=1, unit='GHz')
        self.coax = rf.media.DefinedGammaZ0(frequency=self.freq)
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/'

    def phase_shifter(self, phase_deg):
        return self.coax.line(d=phase_deg, unit='deg')

    def hybrid(self, name='hybrid'):
        Sc = 1/np.sqrt(2)*np.array([[ 0,  1, 1j,  0],
                                    [ 1,  0,  0, 1j],
                                    [1j,  0,  0,  1],
                                    [ 0, 1j,  1,  0]])
        hybrid = rf.Network(frequency=self.freq, s=Sc, name=name)
        return hybrid

    def variable_coupler_network_from_connect(self, phase_deg):
        ps = self.phase_shifter(phase_deg)
        hybrid1, hybrid2 = self.hybrid(), self.hybrid()
        # connecting the networks together.
        # NB: in order to know which port corresponds to which, it is usefull
        # to define different port characteristic impedances like
        # hybrid1.z0 = [11, 12, 13, 14]
        # hybrid2.z0 = [21, 22, 23, 24]
        # ps.z0 = [31, 32]
        # and to make a drawing of each steps.
        # This is not convenient, that's why the Circuit approach can be easier.
        _temp = rf.connect(hybrid1, 2, ps, 0)
        _temp = rf.connect(_temp, 2, hybrid2, 0)
        _temp = rf.innerconnect(_temp, 1, 5)
        # re-order port numbers to match the example
        _temp.renumber([0, 1, 2, 3], [3, 0, 2, 1])
        return _temp

    def variable_coupler_circuit(self, phase_deg):
        ps = self.phase_shifter(phase_deg)
        ps.name = 'ps'  # do not forget the name of the network !
        hybrid1, hybrid2 = self.hybrid('hybrid1'), self.hybrid('hybrid2')

        port1 = rf.Circuit.Port(ps.frequency, 'port1')
        port2 = rf.Circuit.Port(ps.frequency, 'port2')
        port3 = rf.Circuit.Port(ps.frequency, 'port3')
        port4 = rf.Circuit.Port(ps.frequency, 'port4')
        # Note that the order of port appearance is important.
        # 1st port to appear in the connection setup will be the 1st port (0),
        # then second to appear the second port (1), etc...
        # There is no constraint for the order of the connections.
        connections = [
                       [(port1, 0), (hybrid1, 3)],
                       [(port2, 0), (hybrid2, 2)],
                       [(port3, 0), (hybrid2, 1)],
                       [(port4, 0), (hybrid1, 0)],
                       [(hybrid1, 2), (ps, 0)],
                       [(hybrid1, 1), (hybrid2, 0)],
                       [(ps, 1), (hybrid2, 3)],
                      ]

        return rf.Circuit(connections)

    def variable_coupler_network_from_circuit(self, phase_deg):
        return self.variable_coupler_circuit(phase_deg).network

    def test_compare_with_network_connect(self):
        '''
        Compare with the S-parameters obtained from Network.connect
        '''
        phase_deg = np.random.randint(low=0, high=180)
        vc_connect = self.variable_coupler_network_from_connect(phase_deg)
        vc_circuit = self.variable_coupler_network_from_circuit(phase_deg)
        assert_array_almost_equal(vc_connect.s, vc_circuit.s)

    def test_compare_with_designer(self):
        '''
        Compare with the S-parameters obtained from ANSYS Designer
        '''
        for phase_angle in [20, 75]:
            vc_designer = rf.Network(os.path.join(self.test_dir, 'designer_variable_coupler_ideal_'+str(phase_angle)+'deg.s4p'))
            vc_circuit = self.variable_coupler_network_from_circuit(phase_angle)
            assert_array_almost_equal(vc_designer.s, vc_circuit.s, decimal=4)

    def test_compare_connect_and_designer(self):
        '''
        Compare S-parameters obtained from ANSYS Designer with Network.connect
        '''
        for phase_angle in [20, 75]:
            vc_designer = rf.Network(os.path.join(self.test_dir, 'designer_variable_coupler_ideal_'+str(phase_angle)+'deg.s4p'))
            vc_connect = self.variable_coupler_network_from_connect(phase_angle)
            assert_array_almost_equal(vc_designer.s, vc_connect.s, decimal=4)


class CircuitTestGraph(unittest.TestCase):
    '''
    Test functionalities linked to graph method, used in particular for plotting
    '''
    def test_is_networkx_available(self):
        'The networkx package should be available to run these tests'
        self.failUnless('networkx' in sys.modules)

    def setUp(self):
        '''
        Dummy Circuit setup

        Setup a circuit which has various interconnections (2 or 3)
        '''
        self.freq = rf.Frequency(start=1, stop=2, npoints=101)

        # dummy components
        self.R = 100
        self.line_resistor = rf.media.DefinedGammaZ0(frequency=self.freq, Z0=self.R)
        resistor1 = self.line_resistor.resistor(self.R, name='resistor1')
        resistor2 = self.line_resistor.resistor(self.R, name='resistor2')
        resistor3 = self.line_resistor.resistor(self.R, name='resistor3')
        port1 = rf.Circuit.Port(self.freq, name='port1')

        # Connection setup
        self.connections = [
                   [(port1, 0), (resistor1, 0), (resistor3, 0)],
                   [(resistor1, 1), (resistor2, 0)],
                   [(resistor2, 1), (resistor3, 1)]
                ]

        self.C = rf.Circuit(self.connections)

    def test_interstection_dict(self):
        inter_dict = self.C.intersections_dict
        # should have 3 intersections
        self.assert_(len(inter_dict) == 3)
        # All intersections should have at least 2 edges
        for it in inter_dict.items():
            k, cnx = it
            self.assert_(len(cnx) >= 2)

    def test_edge_labels(self):
        edge_labels = self.C.edge_labels
        self.assert_(len(edge_labels) == 7)


class CircuitTestComplexCharacteristicImpedance(unittest.TestCase):
    '''
    Test creating circuits with real and complex port charac.impedances
    '''
    def setUp(self):
        self.f0 = rf.Frequency(75.8, npoints=1, unit='GHz')
        # initial s-param values of A 2 ports network
        self.s0 = np.array([  # dummy values 
            [-0.1000 -0.2000j, -0.3000 +0.4000j],
            [-0.3000 +0.4000j, 0.5000 -0.6000j]]).reshape(-1,2,2)

        # build initial network (under z0=50)
        self.ntw0 = rf.Network(frequency=self.f0, s=self.s0, z0=50, name='dut')

        # complex characteristic impedance to renormalize to
        self.zdut = 100 + 10j
        
        # reference solutions obtained from ANSYS Circuit or ADS (same res) 
        # when z0=[50, zdut]
        self.z_ref = np.array([  # not affected by z0
            [18.0000 -16.0000j, 20.0000 +40.0000j],
            [20.0000 +40.0000j, 10.0000 -80.0000j]]).reshape(-1,2,2)

        self.y_ref = np.array([  # not affected by z0
            [0.0251 +0.0023j, 0.0123 -0.0066j],
            [0.0123 -0.0066j, 0.0052 +0.0055j]]).reshape(-1,2,2)
        
        self.s_ref = np.array([  # renormalized s (power-waves)
            [-0.1374 -0.2957j, -0.1995 +0.5340j],
            [-0.1995 +0.5340j, -0.0464 -0.7006j]]).reshape(-1,2,2)

        # Creating equivalent reference circuit
        port1 = rf.Circuit.Port(self.f0, z0=50, name='port1')
        port2 = rf.Circuit.Port(self.f0, z0=50, name='port2')
        ntw0 = rf.Network(frequency=self.f0, z0=50, s=self.s0, name='dut')

        cnx = [  # z0=[50,50]
            [(port1, 0), (ntw0, 0)],
            [(ntw0, 1), (port2, 0)]
            ]
        self.cir = rf.Circuit(cnx)

        # Creating equivalent circuit with z0 real
        port2_real = rf.Circuit.Port(self.f0, z0=100, name='port2')
        cnx_real = [  # z0=[50,100]
            [(port1, 0), (ntw0, 0)],
            [(ntw0, 1), (port2_real, 0)]
            ]
        self.cir_real = rf.Circuit(cnx_real)

        # Creating equivalent circuit with z0 complex
        port2_complex = rf.Circuit.Port(self.f0, z0=self.zdut, name='port2')
        cnx_complex = [  # z0=[50,zdut]
            [(port1, 0), (ntw0, 0)],
            [(ntw0, 1), (port2_complex, 0)]
            ]
        self.cir_complex = rf.Circuit(cnx_complex)

        # references for each s-param definition
        self.s_legacy = rf.renormalize_s(self.s0, [50,50], [50,self.zdut],
                                         s_def='traveling')
        self.s_power = rf.renormalize_s(self.s0, [50,50], [50,self.zdut],
                                         s_def='power')
        self.s_pseudo = rf.renormalize_s(self.s0, [50,50], [50,self.zdut],
                                         s_def='pseudo')
        # for real values, should be = whatever s-param definition
        self.s_real = rf.renormalize_s(self.s0, [50,50], [50,100])

    def test_verify_reference(self):
        ' Check that the reference results comes from power-waves definition'
        np.testing.assert_allclose(self.s_ref, self.s_power, atol=1e-4)

    def test_reference_s(self):
        ' Check reference z0 circuit, just in case '
        np.testing.assert_allclose(self.cir.network.s, self.s0, atol=1e-4)

    def test_real_z0_s(self):
        ' Check real z0 circuit '
        np.testing.assert_allclose(self.cir_real.network.s, self.s_real, atol=1e-4)

    def test_complexz0_s_vs_legacy(self):
        ' Check complex z0 circuit vs legacy renormalization '
        np.testing.assert_allclose(self.cir_complex.network.s, self.s_legacy, atol=1e-4)

    @unittest.expectedFailure
    def test_complexz0_s_vs_powerwaves(self):
        ' Check complex z0 circuit vs power-waves renormalization '
        np.testing.assert_allclose(self.cir_complex.network.s, self.s_ref, atol=1e-4)
        np.testing.assert_allclose(self.cir_complex.network.s, self.s_power, atol=1e-4)

    @unittest.expectedFailure
    def test_complexz0_s_vs_powerwaves(self):
        ' Check complex z0 circuit vs pseudo-waves renormalization '
        np.testing.assert_allclose(self.cir_complex.network.s, self.s_pseudo, atol=1e-4)

class CircuitTestVoltagesCurrents(unittest.TestCase):
    def setUp(self):
        # setup a test transmission line randomly excited
        self.P_f = np.random.rand()  # forward power in Watt
        self.phase_f = np.random.rand()  # forward phase in rad
        self.Z = np.random.rand()  # source internal impedance, line characteristic impedance and load impedance
        self.L = np.random.rand()  # line length in [m]
        self.freq = rf.Frequency(1, 10, 10, unit='GHz')
        self.line_media = rf.media.DefinedGammaZ0(self.freq, z0=self.Z)  # lossless line medium
        self.line = self.line_media.line(d=self.L, unit='m', name='line')  # transmisison line Network
        
        # forward voltages and currents at the input of the test line
        self.V_in = np.sqrt(2*self.Z*self.P_f)*np.exp(1j*self.phase_f)
        self.I_in = np.sqrt(2*self.P_f/self.Z)*np.exp(1j*self.phase_f)
        # forward voltages and currents at the output of the test line
        theta = rf.theta(self.line_media.gamma, self.freq.f, self.L)  # electrical length 
        self.V_out, self.I_out = rf.tlineFunctions.voltage_current_propagation(self.V_in, self.I_in, self.Z, theta)

        # Equivalent model with Circuit 
        port1 = rf.Circuit.Port(frequency=self.freq, name='port1', z0=self.Z)
        port2 = rf.Circuit.Port(frequency=self.freq, name='port2', z0=self.Z)
        cnx = [
            [(port1, 0), (self.line, 0)],
            [(port2, 0), (self.line, 1)]
        ]
        self.crt = rf.Circuit(cnx)
        # power and phase arrays for Circuit.voltages() and currents()
        self.power = [self.P_f, 0]
        self.phase = [self.phase_f, 0]
        
    def test_tline_voltages(self):
        ' Test voltages for a simple transmission line '
        V_ports = self.crt.voltages_external(self.power, self.phase)
        
        np.testing.assert_allclose(self.V_in, V_ports[:,0])
        np.testing.assert_allclose(self.V_out, V_ports[:,1])
        
    def test_tline_currents(self):
        ' Test currents for a simple transmission line '
        I_ports = self.crt.currents_external(self.power, self.phase)
        
        np.testing.assert_allclose(self.I_in, I_ports[:,0])
        # output current is * -1 as Circuit definition is opposite
        # (toward the Circuit's Port)
        np.testing.assert_allclose(self.I_out, -1*I_ports[:,1])
        

if __name__ == "__main__":
    # Launch all tests
    run_module_suite()



'''
.. currentmodule:: skrf.mathFunctions
=============================================
mathFunctions (:mod:`skrf.mathFunctions`)
=============================================


Provides commonly used mathematical functions.

Complex Component Conversion
---------------------------------
.. autosummary::
        :toctree: generated/

        complex_2_reim
        complex_2_magnitude
        complex_2_db
        complex_2_radian
        complex_2_degree
        complex_2_magnitude

Phase Unwrapping
--------------------------------
.. autosummary::
        :toctree: generated/

        unwrap_rad
        sqrt_phase_unwrap



Unit Conversion
--------------------------------
.. autosummary::
        :toctree: generated/

        radian_2_degree
        degree_2_radian
        np_2_db
        db_2_np

Scalar-Complex Conversion
---------------------------------
These conversions are useful for wrapping other functions that don't
support complex numbers.

.. autosummary::
        :toctree: generated/

        complex2Scalar
        scalar2Complex


Special Functions
---------------------------------
.. autosummary::
        :toctree: generated/

        dirac_delta
        neuman
        null

'''
import numpy as npy
from numpy import pi,angle,unwrap, real, imag, array
from scipy import signal
from scipy.interpolate import interp1d

global LOG_OF_NEG
LOG_OF_NEG = -100

global INF
INF = 1e99

global ALMOST_ZERO
ALMOST_ZERO = 1e-12

## simple conversions
def complex_2_magnitude(input):
    '''
    returns the magnitude of a complex number.
    '''
    return abs(input)

def complex_2_db(input):
    '''
    returns the magnitude in dB of a complex number.

    returns:
            20*log10(|z|)
    where z is a complex number
    '''
    return magnitude_2_db(npy.abs( input))

def complex_2_db10(input):
    '''
    returns the magnitude in dB of a complex number.

    returns:
            10*log10(|z|)
    where z is a complex number
    '''
    return mag_2_db10(npy.abs( input))


def complex_2_radian(input):
    '''
    returns the angle complex number in radians.

    '''
    return npy.angle(input)

def complex_2_degree(input):
    '''
    returns the angle complex number in radians.

    '''
    return npy.angle(input, deg=True)

def complex_2_quadrature(z):
    '''
    takes a complex number and returns quadrature, which is (length, arc-length from real axis)
    '''
    return ( npy.abs(z), npy.angle(z)*npy.abs(z))

def complex_components(z):
    '''
    break up a complex array into all possible scalar components

    takes: complex ndarray
    return:
            c_real: real part
            c_imag: imaginary part
            c_angle: angle in degrees
            c_mag:  magnitude
            c_arc:  arclength from real axis, angle*magnitude
    '''
    return (npy.real(z), npy.imag(z), npy.angle(z,deg=True), complex_2_quadrature(z)[0], complex_2_quadrature(z)[1])

def complex_2_reim(z):
    '''
    returns real and imaginary parts of a complex number.
    
    takes:
            input: complex number or array
    return:
            real: real part of input
            imag: imaginary part of input


    note: this just calls 'complex_components'
    '''
    out = complex_components(z)
    return (out[0],out[1])

def magnitude_2_db(input,zero_nan=True):
    '''
    converts linear magnitude to db

     db is given by
            20*log10(|z|)
    where z is a complex number
    '''
    if zero_nan:
        out = 20 * npy.log10(input)
        try:
            out[npy.isnan(out)] = LOG_OF_NEG
        except (TypeError):
            # input is a number not array-like
            if npy.isnan(out):
                return LOG_OF_NEG

    else:
        out = 20*npy.log10(input)

    return out

mag_2_db = magnitude_2_db

def mag_2_db10(input,zero_nan=True):
    '''
    converts linear magnitude to db

     db is given by
            10*log10(|z|)
    where z is a complex number
    '''
    out = 10 * npy.log10(input)
    if zero_nan:
        try:
            out[npy.isnan(out)] = LOG_OF_NEG
        except (TypeError):
            # input is a number not array-like
            if npy.isnan(out):
                return LOG_OF_NEG


    return out

def db_2_magnitude(input):
    '''
    converts dB to linear magnitude.

    returns:
            10**((z)/20.)
    where z is a complex number
    '''
    return 10**((input)/20.)

db_2_mag = db_2_magnitude


def db10_2_mag(input):
    '''
    converts dB to linear magnitude

    returns:
            10**((z)/10.)
    where z is a complex number
    '''
    return 10**((input)/10.)


def magdeg_2_reim(mag,deg):
    '''
    converts linear magnitude and phase (in deg) arrays into a complex array
    '''
    return mag*npy.exp(1j*deg*pi/180.)

def dbdeg_2_reim(db,deg):
    '''
    converts dB magnitude and phase (in deg) arrays into a complex array
    '''
    return magdeg_2_reim(db_2_magnitude(db),deg)


def db_2_np(x):
    '''
    Converts a value in decibel (dB) to neper (Np).
    '''
    return (npy.log(10)/20) * x


def np_2_db(x):
    '''
    Converts a value in Nepers (Np) to decibel (dB).
    '''
    return 20/npy.log(10) * x

def radian_2_degree(rad):
    return (rad)*180/pi

def degree_2_radian(deg):
    return (deg)*pi/180.

def feet_2_meter(feet=1):
    '''
    Convert length in feet to meter.
    
    1 foot is equal to 0.3048 meters:
        
    Parameters
    ----------
    feet : number or array-like
        length in feet

    Returns
    -------
    meter: number or array-like
        length in meter

    '''
    return 0.3048*feet

def meter_2_feet(meter=1):
    '''
    Convert length in meter to feet.
    
    1 meter is equal to 0.3.28084 feet:
            
    Parameters
    ----------
    meter : number or array-like
        length in meter

    Returns
    -------
    feet : number or array-like
        length in feet

    '''
    return 3.28084*meter


def db_per_100feet_2_db_per_100meter(db_per_100feet=1):
    '''
    Convert attenuation values given in dB/100ft to dB/100m.
    
    db_per_100meter = db_per_100feet * rf.meter_2_feet()
    
    Parameters
    ----------
    db_per_100feet : number or array-like
        Attenuation in dB/ 100 ft

    Returns
    -------
    db_per_100meter : number or array-like
        Attenuation in dB/ 100 m

    '''
    return db_per_100feet * 100 / feet_2_meter(100)

def unwrap_rad(input):
    '''
    unwraps a phase given in radians

    '''
    return unwrap(input,axis=0)

def sqrt_known_sign(z_squared, z_approx):
    '''
    Returns sqrt of complex number, with sign chosen to match `z_approx`

    Parameters
    -------------
    z_squared : number, array-like
        the complex to be square-rooted
    z_approx : number, array-like
        the approximate value of z. sign of z is chosen to match that of
        z_approx

    Returns
    ----------
    z : number, array-like (same type as z_squared)
        square root of z_squared.



    '''
    z = npy.sqrt(z_squared)
    return npy.where(
        npy.sign(npy.angle(z)) == npy.sign(npy.angle(z_approx)),
        z, z.conj())

def find_correct_sign(z1,z2,z_approx):
    '''
    Create new vector from z1, z2 choosing elements with sign matching z_approx

    This is used when you have to make a root choice on a complex number.
    and you know the approximate value of the root.

    .. math::

        z1,z2 = \\pm \\sqrt(z^2)


    Parameters
    ------------
    z1 : array-like
        root 1
    z2 : array-like
        root 2
    z_approx : array-like
        approximate answer of z

    Returns
    ----------
    z3 : npy.array
        array built from z1 and z2 by
        z1 where sign(z1) == sign(z_approx), z2 else

    '''
    return npy.where(
    npy.sign(npy.angle(z1)) == npy.sign(npy.angle(z_approx)),z1, z2)

def find_closest(z1,z2,z_approx):
    '''
    Returns z1 or z2  depending on which is  closer to z_approx


    Parameters
    ------------
    z1 : array-like
        root 1
    z2 : array-like
        root 2
    z_approx : array-like
        approximate answer of z

    Returns
    ----------
    z3 : npy.array
        array built from z1 and z2

    '''
    z1_dist = abs(z1-z_approx)
    z2_dist = abs(z2-z_approx)

    return npy.where(z1_dist<z2_dist,z1, z2)

def sqrt_phase_unwrap(input):
    '''
    takes the square root of a complex number with unwrapped phase

    this idea came from Lihan Chen
    '''
    return npy.sqrt(abs(input))*\
            npy.exp(0.5*1j*unwrap_rad(complex_2_radian(input)))


# mathematical functions
def dirac_delta(x):
    '''
    the Dirac function.

    can take numpy arrays or numbers
    returns 1 or 0 '''
    return (x==0)*1.+(x!=0)*0.
def neuman(x):
    '''
    neumans number

    2-dirac_delta(x)

    '''
    return 2. - dirac_delta(x)
def null(A, eps=1e-15):
    '''
     calculates the null space of matrix A.
    i found this on stack overflow.
     '''
    u, s, vh = npy.linalg.svd(A)
    null_space = npy.compress(s <= eps, vh, axis=0)
    return null_space.T


def inf_to_num(x):
    '''
    converts inf and -inf's to large numbers

    Parameters
    ------------
    x : array-like or number
        the input array or number
    Returns
    -------
    '''
    #TODO: make this valid for complex arrays
    try:
        x[npy.isposinf(x)] = INF
        x[npy.isneginf(x)] = -1*INF

    except(TypeError):
        x = npy.array(x)
        x[npy.isposinf(x)] = INF
        x[npy.isneginf(x)] = -1*INF


def cross_ratio(a,b,c,d):
    '''
    The cross ratio


    defined as

    .. math::

        \frac{(a-b)(c-d)}{(a-d)*(c-b)}


    Parameters
    -------------
    a,b,c,d : complex numbers, or arrays
        mm


    '''
    return ((a-b)*(c-d))/((a-d)*(c-b))




    

def complexify(f, name=None):
    '''
    make  f(scalar)  into f(complex)
    
    if the real/imag arguments are not first, then you may specify the
    name given to them as kwargs.
    '''
    
        
    def f_c(z, *args, **kw):
        if name is not None:
            kw_re= {name:real(z)}
            kw_im= {name:imag(z)}
            kw_re.update(kw)
            kw_im.update(kw)
            return f(*args, **kw_re)+ 1j*f(*args, **kw_im)
        else:
            return f(real(z), *args,**kw)+ 1j*f(imag(z), *args, **kw)
    return f_c



# old functions just for reference
def complex2Scalar(input):
    '''
    Serializes a list/arary of complex numbers
    
    
    produces the following output for input list `x`
    
    x[0].real, x[0].imag, x[1].real, x[1].imag, etc
    '''
    input= npy.array(input)
    output = []
    for k in input:
        output.append(npy.real(k))
        output.append(npy.imag(k))
    return npy.array(output).flatten()

def scalar2Complex(input):
    '''
    inverse of `complex2Scalar`
    '''
    input= npy.array(input)
    output = []

    for k in range(0,len(input),2):
        output.append(input[k] + 1j*input[k+1])
    return npy.array(output).flatten()

def complex2dB(complx):
    dB = 20 * npy.log10(npy.abs( (npy.real(complx) + 1j*npy.imag(complx) )))
    return dB

def flatten_c_mat(s, order ='F'):
    '''
    take a 2D (mxn) complex matrix and serialize and flatten it
    
    by default (using order='F') this generates the following 
    from a  2x2 
    
    [s11,s12;s21,s22]->[s11re,s11im,s21re,s12im, ...]
    
    Parameters
    ------------
    s : ndarray
        input 2D array 
    order : ['F','C']
        order of flattening
    '''
    return complex2Scalar(s.flatten(order='F'))


def complex2ReIm(complx):
    return npy.real(complx), npy.imag(complx)

def complex2MagPhase(complx,deg=False):
    return npy.abs(complx), npy.angle(complx,deg=deg)

def rand_c(*args):
    '''
    Creates a complex random array of shape s.

    The bounds on real and imaginary values are (-1,1)


    Parameters
    -----------
    s : list-like
        shape of array

    Examples
    ---------
    >>> x = rf.rand_c(2,2)
    '''
    s = npy.array(args)
    return 1-2*npy.random.rand(npy.product(s)).reshape(s) + \
        1j-2j*npy.random.rand(npy.product(s)).reshape(s)



def psd2TimeDomain(f,y, windowType='hamming'):
    '''convert a one sided complex spectrum into a real time-signal.
    takes
            f: frequency array,
            y: complex PSD array
            windowType: windowing function, defaults to rect

    returns in the form:
            [timeVector, signalVector]
    timeVector is in inverse units of the input variable f,
    if spectrum is not baseband then, timeSignal is modulated by
            exp(t*2*pi*f[0])
    so keep in mind units, also due to this f must be increasing left to right'''


    # apply window function
    #TODO: make sure windowType exists in scipy.signal
    if (windowType != 'rect' ):
        exec("window = signal.%s(%i)" % (windowType,len(f)))
        y = y * window

    #create other half of spectrum
    spectrum = (npy.hstack([npy.real(y[:0:-1]),npy.real(y)])) + \
            1j*(npy.hstack([-npy.imag(y[:0:-1]),npy.imag(y)]))

    # do the transform
    df = abs(f[1]-f[0])
    T = 1./df
    timeVector = npy.linspace(-T/2.,T/2,2*len(f)-1)
    signalVector = npy.fft.ifftshift(npy.fft.ifft(npy.fft.ifftshift(spectrum)))

    #the imaginary part of this signal should be from fft errors only,
    signalVector= npy.real(signalVector)
    # the response of frequency shifting is
    # exp(1j*2*pi*timeVector*f[0])
    # but I would have to manually undo this for the inverse, which is just
    # another variable to require. The reason you need this is because
    # you can't transform to a bandpass signal, only a lowpass.
    #
    return timeVector, signalVector

def rational_interp(x, y, d=4, epsilon=1e-9, axis=0):
    """
    Interpolates function using rational polynomials of degree `d`.

    Interpolating function is singular when xi is exactly one of the
    original x points. If xi is closer than epsilon to one of the original points,
    then the value at that points is returned instead.

    Implementation is based on [0]_.

    References
    ------------
    .. [0] M. S. Floater and K. Hormann, "Barycentric rational interpolation with no poles and high rates of approximation," Numer. Math., vol. 107, no. 2, pp. 315-331, Aug. 2007
    """
    n = len(x)
    w = npy.zeros(n)
    for k in range(n):
        for i in range(max(0,k-d), min(k+1, n-d)):
            p = 1.0
            for j in range(i,min(n,i+d+1)):
                if j == k:
                    continue
                p *= 1/(x[k] - x[j])
            w[k] += ((-1)**i)*p

    if axis != 0:
        raise NotImplementedError("Axis other than 0 is not implemented")

    def fx(xi):
        def find_nearest(a, values, epsilon):
            idx = npy.abs(npy.subtract.outer(a, values)).argmin(0)
            return npy.abs(a[idx] - values) < epsilon

        def find_nearest_value(a, values, y):
            idx = npy.abs(npy.subtract.outer(a, values)).argmin(0)
            return y[idx]

        nearest = find_nearest(x, xi, epsilon)
        nearest_value = find_nearest_value(x, xi, y)

        #There needs to be a cleaner way
        w_shape = [1]*len(y.shape)
        w_shape[0] = -1
        wr = w.reshape(*w_shape)

        with npy.errstate(divide='ignore', invalid='ignore'):
            #nans will be fixed later
            v = npy.sum([y[i]*wr[i]/((xi - x[i]).reshape(*w_shape)) for i in range(n)], axis=0)\
                /npy.sum([w[i]/((xi - x[i]).reshape(*w_shape)) for i in range(n)], axis=0)

        for e, i in enumerate(nearest):
            if i:
                v[e] = nearest_value[e]

        return v

    return fx
    
def ifft(x):
    """
    Transforms S-parameters to time-domain bandpass.
    """
    return npy.fft.fftshift(npy.fft.ifft(x, axis=0), axes=0)

def irfft(x, n=None):
    """
    Transforms S-parameters to time-domain, assuming complex conjugates for
    values corresponding to negative frequencies.
    """
    return npy.fft.fftshift(npy.fft.irfft(x, axis=0, n=n), axes=0)


# Matrix functions

def is_square(mat):
    """
    Tests whether mat is a square matrix

    Parameters
    ----------
    mat : npy.matrix
        Matrix to test for being square
    """
    return mat.shape[0] == mat.shape[1]


def is_unitary(mat, tol=ALMOST_ZERO):
    """
    Tests mat for unitariness

    Parameters
    ----------
    mat : npy.matrix
        Matrix to test for unitariness
    tol : float
        Absolute tolerance
    """
    if not is_square(mat):
        return False
    return npy.allclose(get_Hermitian_transpose(mat) * mat,
                        npy.identity(mat.shape[0]), atol=tol)


def is_symmetric(mat, tol=ALMOST_ZERO):
    """
    Tests mat for symmetry

    Parameters
    ----------
    mat : npy.matrix
        Matrix to test for symmetry
    tol : float
        Absolute tolerance
    """
    if not is_square(mat):
        return False
    return npy.allclose(mat, mat.transpose(), atol=tol)


def get_Hermitian_transpose(mat):
    """
    Returns the conjugate transpose of mat

    Parameters
    ----------
    mat : npy.matrix
        Matrix to compute the conjugate transpose of
    """
    return mat.transpose().conjugate()


def is_Hermitian(mat, tol=ALMOST_ZERO):
    """
    Tests whether mat is Hermitian

    Parameters
    ----------
    mat : npy.matrix
        Matrix to test for being Hermitian
    tol : float
        Absolute tolerance
    """
    if not is_square(mat):
        return False
    return npy.allclose(mat, get_Hermitian_transpose(mat), atol=tol)


def is_positive_definite(mat, tol=ALMOST_ZERO):
    """
    Tests mat for positive definiteness by verifying that
    (1) mat is symmetric
    (2) it's possible to compute the Cholesky decomposition of mat.

    Parameters
    ----------
    mat : npy.matrix
        Matrix to test for positive definiteness
    tol : float
        Absolute tolerance
    """
    if not is_Hermitian(mat, tol=tol):
        return False
    try:
        npy.linalg.cholesky(mat)
        return True
    except npy.linalg.LinAlgError:
        return False


def is_positive_semidefinite(mat, tol=ALMOST_ZERO):
    """
    Tests mat for positive semidefiniteness by checking
    whether all eigenvalues of mat are nonnegative within a certain tolerance

    Parameters
    ----------
    mat : npy.matrix
        Matrix to test for positive semidefiniteness
    tol : float
        Absolute tolerance in determining nonnegativity due to loss of precision
        when computing the eigenvalues of mat
    """
    if not is_Hermitian(mat):
        return False
    try:
        v = npy.linalg.eigvalsh(mat)
    except npy.linalg.LinAlgError:
        return False
    return npy.all(v > -tol)




import unittest
import os
import numpy as npy
from pathlib import Path
from zipfile import ZipFile

import skrf as rf
from skrf.io.touchstone import Touchstone


class TouchstoneTestCase(unittest.TestCase):
    '''
    TouchstoneTestCase tests the IO of Touchstone files
    '''
    def setUp(self):
        '''
        Sets up the test directory
        '''
        self.test_dir = os.path.dirname(os.path.abspath(__file__))+'/'

    def test_read_data(self):
        '''
        This test reads data from simple_touchstone.s2p and compares with known
        true values.
        '''
        filename = os.path.join(self.test_dir, 'simple_touchstone.s2p')
        touch = Touchstone(filename)
        f, s = touch.get_sparameter_arrays()
        z0 = complex(touch.resistance)
        f_true = npy.array([1.00000000e+09, 1.10000000e+09])
        s_true = npy.array([[[1.+2.j, 5.+6.j], [3.+4.j, 7.+8.j]],
                            [[9.+10.j, 13.+14.j], [11.+12.j, 15.+16.j]]])
        z0_true = 50+50j

        self.assertTrue((f == f_true).all())
        self.assertTrue((s == s_true).all())
        self.assertTrue((z0 == z0_true))


    def test_read_from_fid(self):
        '''
        This tests reading touch stone data from a file object as compared with
        a string path and name of the file.
        '''
        with open(os.path.join(self.test_dir, 'simple_touchstone.s2p')) as fid:
            touch = Touchstone(fid)
        f, s = touch.get_sparameter_arrays()
        z0 = complex(touch.resistance)
        f_true = npy.array([1.00000000e+09, 1.10000000e+09])
        s_true = npy.array([[[1.+2.j, 5.+6.j], [3.+4.j, 7.+8.j]],
                            [[9.+10.j, 13.+14.j], [11.+12.j, 15.+16.j]]])
        z0_true = 50+50j

        self.assertTrue((f == f_true).all())
        self.assertTrue((s == s_true).all())
        self.assertTrue((z0 == z0_true))

    def test_get_sparameter_data(self):
        '''
        This tests the get_sparameter_data function.

        '''
        with open(os.path.join(self.test_dir, 'simple_touchstone.s2p')) as fid:
            touch = Touchstone(fid)

        expected_keys = ["frequency", "S11R", "S11I", "S12R", "S12I",
                "S21R", "S21I", "S22R", "S22I", ]

        unexpected_keys = ['S11DB', 'S11M', ]

        # get dict data structure
        sp_ri = touch.get_sparameter_data(format="ri")

        # test data structure
        for ek in expected_keys:
            self.assertTrue(ek in sp_ri)

        for uk in unexpected_keys:
            self.assertFalse(uk in sp_ri)

        # test data contents
        expected_sp_ri = {
            'frequency': npy.array([1.0e+09, 1.1e+09]),
            'S11R': npy.array([1., 9.]),
            'S11I': npy.array([ 2., 10.]),
            'S21R': npy.array([ 3., 11.]),
            'S21I': npy.array([ 4., 12.]),
            'S12R': npy.array([ 5., 13.]),
            'S12I': npy.array([ 6., 14.]),
            'S22R': npy.array([ 7., 15.]),
            'S22I': npy.array([ 8., 16.]),
        }

        for k in sp_ri:
            self.assertTrue(k in expected_sp_ri)

            self.assertTrue( (expected_sp_ri[k] == sp_ri[k]).all(),
                    msg='Field %s does not match. Expected "%s", got "%s"'%(
                        k, str(expected_sp_ri[k]), str(sp_ri[k]))  )


    def test_HFSS_touchstone_files(self):
        """ 
        HFSS can export additional information in the Touchstone file
        such as gamma and z0 for each port. However, the way there are stored
        depend of the HFSS version... 
        
        In versions before 2020 R2, data were stored as following:
        
        ! Gamma ! re1 im2 re2 im2 re3 im3 re4 im4  
        !       re5 im5 re6 im6 re7 im7 re8 im8
        !       re9 im9  [etc]
        ! Port Impedancere1 im2 re2 im2 re3 im3 re4 im4  
        !       re5 im5 re6 im6 re7 im7 re8 im8
        !       re9 im9  [etc]
            [NB: there is an extra ! before re1 for Gamma]
            [NB: re1 value is stuck to the 'e' of Impedance]
        
        Since version 2020 R2n the data are stored in a single line:
        
        ! Gamma re1 im2 re2 im2 re3 im3 re4 im4 re5 im5 re6 im6 re7 im7 re8 im8 [etc]
        ! Port Impedance re1 im2 re2 im2 re3 im3 re4 im4 re5 im5 re6 im6 re7 im7 re8 im8 [etc]
            [NB: re1 value is no more stuck to the 'e' of Impedance]
        
        This test checks that the shape of gamma and z0 matche the rank of the Network 
        for Touchstone files of various port obtained from different HFSS version
        """
        HFSS_RELEASES= ['HFSS_2019R2', 'HFSS_2020R2']

        p = Path('.')
        for hfss_release in HFSS_RELEASES:
            for sNp_file in p.glob(hfss_release+'/*.s*'):
                touchst = Touchstone(sNp_file.as_posix())
                gamma, z0 = touchst.get_gamma_z0()
                
                assert(gamma.shape[-1] == touchst.rank)
                assert(z0.shape[-1] == touchst.rank)

suite = unittest.TestLoader().loadTestsFromTestCase(TouchstoneTestCase)
unittest.TextTestRunner(verbosity=2).run(suite)






'''

.. module:: skrf.vi.stages
++++++++++++++++++++++++++++++++++++++++++++++++++++
Stages  (:mod:`skrf.vi.stages`)
++++++++++++++++++++++++++++++++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ESP300

'''
from time import sleep
import numpy as npy

from . ivihack import Driver


class ESP300(Driver):
    '''
    Newport Universal Motion Controller/Driver Model ESP300

    all axis control commands are sent to the number axis given by the
    local variable self.current_axis. An example usage ::

        from skrf.vi.stages import ESP300
        esp = ESP300()
        esp.current_axis = 1
        esp.position = 10
        print(esp.position)
    '''
    UNIT_DICT = {\
            'enoder count':0,\
            'motor step':1,\
            'millimeter':2,\
            'micrometer':3,\
            'inches':4,\
            'milli inches':5,\
            'micro inches':6,\
            'degree':7,\
            'gradient':8,\
            'radian':9,\
            'milliradian':10,\
            'microradian':11,\
            }

    def __init__(self, address=1, current_axis=1,\
            always_wait_for_stop=True,delay=0,**kwargs):
        '''
        Initializer

        Parameters
        -------------
        address :   int
            Gpib address
        current_axis :   int
            number of current axis
        always_wait_for_stop :   Boolean
            wait for stage to stop before
            returning control to calling program
        \*\*kwargs :
            passed to GpibInstrument initializer
        '''

        if isinstance(address,int):
            resource = 'GPIB::%i::INSTR'%address
        else:
            resource = address

        Driver.__init__(self,resource = resource, **kwargs)
        self.current_axis = current_axis
        self.always_wait_for_stop = always_wait_for_stop
        self.delay=delay
    @property
    def current_axis(self):
        '''
        current axis used in all subsequent commands
        '''
        return self._current_axis
    @current_axis.setter
    def current_axis(self, input):
        '''
        takes:
                input:  desired current axis number, int []
        '''
        self._current_axis = input


    @property
    def velocity(self):
        '''
        the velocity of current axis
        '''
        command_string = 'VA'
        return (float(self.ask('%i%s?'%(self.current_axis,command_string))))
    @velocity.setter
    def velocity(self,input):
        command_string = 'VA'
        self.write('%i%s%f'%(self.current_axis,command_string,input))

    @property
    def acceleration(self):
        command_string = 'AC'
        return (self.ask('%i%s?'%(self.current_axis,command_string)))
    @acceleration.setter
    def acceleration(self,input):
        command_string = 'AC'
        self.write('%i%s%f'%(self.current_axis,command_string,input))

    @property
    def deceleration(self):
        command_string = 'AG'
        return (self.ask('%i%s?'%(self.current_axis,command_string)))
    @deceleration.setter
    def deceleration(self,input):
        command_string = 'AG'
        self.write('%i%s%f'%(self.current_axis,command_string,input))

    @property
    def position_relative(self):
        raise NotImplementedError('See position property for reading position')
    @position_relative.setter
    def position_relative(self,input):
        command_string = 'PR'
        self.write('%i%s%f'%(self.current_axis,command_string,input))
        if self.always_wait_for_stop:
            self.wait_for_stop()

    @property
    def position(self):
        command_string = 'TP'
        return float(self.ask('%i%s'%(self.current_axis,command_string)))
    @position.setter
    def position(self,input):
        '''
        set the position of current axis to input
        '''
        command_string = 'PA'
        self.write('%i%s%f'%(self.current_axis,command_string,input))
        if self.always_wait_for_stop:
            self.wait_for_stop()
    @property
    def home(self):
        raise NotImplementedError
    @home.setter
    def home(self, input):
        command_string = 'DH'
        self.write('%i%s%f'%(self.current_axis,command_string,input))


    @property
    def units(self):
        raise NotImplementedError('I dont know how to read units')
    @units.setter
    def units(self, input):
        '''
         set axis units for all commands.
         takes:
                input: a string, describing the units here are a list of
                        possibilities.
                        'encoder count'
                        'motor step'
                        'millimeter'
                        'micrometer'
                        'inches'
                        'milli inches'
                        'micro inches'
                        'degree'
                        'gradient'
                        'radian'
                        'milliradian'
                        'microradian'
        '''
        command_string = 'SN'
        self.write('%i%s%i'%(self.current_axis,command_string, self.UNIT_DICT[input]))

    @property
    def error_message(self):
        return (self.ask('TB?'))

    @property
    def motor_on(self):
        command_string = 'MO'
        return (self.ask('%i%s?'%(self.current_axis,command_string)))
    @motor_on.setter
    def motor_on(self,input):
        if input:
            command_string = 'MO'
            self.write('%i%s'%(self.current_axis,command_string))
        if not input:
            command_string = 'MF'
            self.write('%i%s'%(self.current_axis,command_string))

    def send_stop(self):
        command_string = 'ST'
        self.write('%i%s'%(self.current_axis,command_string))

    def wait_for_stop(self):
        command_string = 'WS'
        self.write('%i%s%i'%(self.current_axis,command_string, 1e3*self.delay))
        tmp = self.position



from traits.api import *
from traitsui.api import *#View, Item, ButtonEditor, Group, HSplit,
from traitsui.menu import *

import pylab as plb
import skrf as rf

from plotTool import PlotTool

class TouchstonePlotter(HasTraits):
    cur_dir = Directory('')
    ntwk_dict = Property(depends_on = 'cur_dir')
    available_ntwk_list = Property(depends_on = 'cur_dir')
    active_ntwk_name = Enum(values='available_ntwk_list')
    active_ntwk = Property(depends_on = 'active_ntwk_name')

    ports = Property(depends_on = 'active_ntwk_name')
    m = Enum(values = 'ports')
    n = Enum(values = 'ports')
    param = Enum('s',['s','z','y','a'])
    form = Enum('Mag (dB)', ['Mag (dB)','Phase (deg)','Smith',
        'Mag (lin)', 'Phase (lin)','Real','Imag'])

    plot_button = Button('Plot ')

    plot_tool = Instance(PlotTool,())

    @cached_property
    def _get_ntwk_dict(self):
        try:
            return rf.ran(self.cur_dir)
        except(OSError):
            return {}

    def _get_available_ntwk_list(self):
        return  self.ntwk_dict.keys()

    def _get_active_ntwk(self):
        try:
            return self.ntwk_dict[self.active_ntwk_name]
        except:
            pass

    def _get_ports(self):
        try:
            return ['All']+range(self.active_ntwk.nports)
        except:
            return []

    def _plot_button_fired(self):
        form_dict = {
            'Mag (dB)': 'db',
            'Mag (lin)': 'mag',
            'Phase (deg)': 'deg',
            'Phase Unwrapped (deg)': 'deg_unwrap',
            'Smith': 'smith',
            'Real': 're',
            'Imag': 'im',
            }
        form = form_dict[self.form]
        m = None if self.m == 'All' else self.m
        n = None if self.n == 'All' else self.n


        self.active_ntwk.__getattribute__(\
            'plot_{}_{}'.format(self.param, form))(m=m, n=n)


        plb.draw();plb.show()

    view = View(
        Item('cur_dir'),
        HSplit(
            Item('active_ntwk_name', show_label=False),
            Item('plot_button', show_label=False),
            Item('param', show_label=False),
            Item('form', show_label=False),
            Item('m'),
            Item('n'),

            ),
        '_',


        Item('plot_tool',style='custom',show_label=False),

        resizable=True,
        width = 800, height = 300,
        title = 'Skrf Touchstone plotter')


if __name__  == '__main__':
    a = TouchstonePlotter()
    a.configure_traits()



# -*- coding: utf-8 -*-
'''
.. module:: skrf.media.coaxial

============================================================
coaxial (:mod:`skrf.media.coaxial`)
============================================================

A coaxial transmission line defined from its electrical or geometrical/physical properties

.. autosummary::
   :toctree: generated/

   Coaxial

'''

#from copy import deepcopy
from scipy.constants import  epsilon_0, mu_0, pi, c
from numpy import sqrt, log, real, imag, exp, size
from ..tlineFunctions import surface_resistivity
from .distributedCircuit import DistributedCircuit
from .media import Media, DefinedGammaZ0
from ..constants import INF
from ..mathFunctions import feet_2_meter, db_per_100feet_2_db_per_100meter, db_2_np


class Coaxial( DistributedCircuit,Media ):
    '''
    A coaxial transmission line defined in terms of its inner/outer
    diameters and permittivity



    Parameters
    ----------
    frequency : :class:`~skrf.frequency.Frequency` object

    z0 : number, array-like, or None
        the port impedance for media. Only needed if  its different
        from the characterisitc impedance of the transmission
        line. if z0 is None then will default to Z0
    Dint : number, or array-like
        inner conductor diameter, in m
    Dout : number, or array-like
        outer conductor diameter, in m
    epsilon_r=1 : number, or array-like
        relative permittivity of the dielectric medium
    tan_delta=0 : number, or array-like
        loss tangent of the dielectric medium
    sigma=infinity : number, or array-like
        conductors electrical conductivity, in S/m


    TODO : different conductivity in case of different conductor kind

    Notes
    ----------
    Dint, Dout, epsilon_r, tan_delta, sigma can all be vectors as long
    as they are the same length

    References
    ---------
    .. [#] Pozar, D.M.; , "Microwave Engineering", Wiley India Pvt. Limited, 1 sept. 2009



        '''
    ## CONSTRUCTOR
    def __init__(self, frequency=None,  z0=None, Dint=.81e-3,
                 Dout=5e-3, epsilon_r=1, tan_delta=0, sigma=INF,
                 *args, **kwargs):



        Media.__init__(self, frequency=frequency,z0=z0)

        self.Dint, self.Dout = Dint,Dout
        self.epsilon_r, self.tan_delta, self.sigma = epsilon_r, tan_delta, sigma
        self.epsilon_prime = epsilon_0*self.epsilon_r
        self.epsilon_second = epsilon_0*self.epsilon_r*self.tan_delta

    @classmethod
    def from_attenuation_VF(cls, frequency=None, z0=None, Z0=50,
                         att=0, unit='dB/m', VF=1):
        """
        Init from electrical properties of the line: attenuation and velocity factor.

        Attenuation can be expressed in dB/m, dB/100m, dB/ft, dB/100ft, Neper/m or Neper/ft.
        Default unit is dB/m. A different unit is set by the `unit` parameter.

        Parameters
        ----------
        frequency : :class:`~skrf.frequency.Frequency` object

        z0 : number, array-like, or None
            the port impedance for media. Only needed if its different
            from the characterisitc impedance of the transmission
            line. if z0 is None then will default to Z0
        Z0 : number
            desired characteristic impedance
        att : number, or array-like. optional
            Attenuation of the coaxial line. The default is 0.
            If passed as an array, should be of same size than the frequency.
        unit : string, optional
            Unit of the attenuation. Can be: 'dB/m', dB/100m'', 'dB/ft', 'dB/100ft',
            'Neper/m' or 'Neper/ft' (or 'Np/m', 'Np/ft'). The default is 'dB/m'.
        VF : number, or array-like. optional
            Velocity Factor VF [VF]_. The default is 1.
            If passed as an array, should be of same size than the frequency.

        References
        ----------
        .. [VF] : https://www.microwaves101.com/encyclopedias/light-phase-and-group-velocities

        """
        # test size of parameters
        if size(att) not in (1, size(frequency)):
            raise ValueError('Attenuation should be scalar or of same size that the frequency.')
    
        # create gamma
        if unit in ('dB/m', 'db/m'):
            alpha = db_2_np(att)
        elif unit in ('dB/100m', 'db/100m'):
            alpha = db_2_np(att/100)
        elif unit in ('dB/ft', 'dB/feet'):
            alpha = db_2_np(att/feet_2_meter())
        elif unit in ('dB/100ft', 'dB/100feet'):
            alpha = db_2_np(db_per_100feet_2_db_per_100meter(att)/100)
        elif unit in ('Np/m', 'np/m', 'n/m'):
            alpha = att
        elif unit in ('Neper/feet', 'Neper/ft',
                      'Np/feet', 'Np/ft',
                      'np/feet', 'np/ft',
                      'N/feet', 'N/ft'):
            alpha = att/feet_2_meter()
        else:
            raise ValueError('Incorrect attenuation unit. Please see documentation. ', unit)

        beta = 2 * pi * frequency.f / c / VF

        gamma = alpha + 1j*beta

        # return media object from z0 and gamma
        return DefinedGammaZ0(frequency=frequency, gamma=gamma,
                                    z0=z0, Z0=Z0)

    @classmethod
    def from_Z0_Dout(cls, frequency=None, z0=None,Z0=50,  epsilon_r=1,
                     Dout=5e-3, **kw):
        '''
        Init from characteristic impedance and outer diameter

        Parameters
        ----------
        frequency : :class:`~skrf.frequency.Frequency` object

        z0 : number, array-like, or None
            the port impedance for media. Only needed if  its different
            from the characterisitc impedance of the transmission
            line. if z0 is None then will default to Z0
        Z0 : number
            desired characteristic impedance
        Dout : number, or array-like
            outer conductor diameter, in m
        epsilon_r=1 : number, or array-like
            relative permittivity of the dielectric medium
        **kw :
            passed to __init__
        '''
        ep= epsilon_0*epsilon_r

        if imag(Z0) !=0:
            raise NotImplementedError()

        b = Dout/2.
        b_over_a = exp(2*pi*Z0*sqrt(ep/mu_0))
        a = b/b_over_a
        Dint = 2*a
        return cls(frequency=frequency, z0 = z0, Dint=Dint, Dout=Dout,
                    epsilon_r=epsilon_r, **kw)


    @property
    def Rs(self):
        '''
        Surface resistivity in Ohm/area

        Returns
        -------
        Rs : number or array
            surface resistivity

        '''
        f  = self.frequency.f
        rho = 1./self.sigma
        mu_r =1
        return surface_resistivity(f=f,rho=rho, mu_r=mu_r)

    @property
    def a(self):
        '''
        Inner radius of the coaxial line

        Returns
        -------
        a : float
            Inner radius

        '''
        return self.Dint/2.

    @property
    def b(self):
        '''
        Outer radius of the coaxial line
        
        Returns
        -------
        b : float
            Outer radius
        '''
        return self.Dout/2.


    # derivation of distributed circuit parameters
    @property
    def R(self):
        '''
        Distributed resistance R, in Ohm/m

        Returns
        -------
        R : number, or array-like
            distributed resistance, in Ohm/m

        '''
        return self.Rs/(2.*pi)*(1./self.a + 1./self.b)

    @property
    def L(self):
        '''
        Distributed inductance L, in H/m

        Returns
        -------
        L : number, or array-like
            distributed inductance, in  H/m

        '''
        return mu_0/(2.*pi)*log(self.b/self.a)

    @property
    def C(self):
        '''
        Distributed capacitance C, in F/m

        Returns
        -------
        C : number, or array-like
            distributed capacitance, in F/m

        '''
        return 2.*pi*self.epsilon_prime/log(self.b/self.a)

    @property
    def G(self):
        '''
        Distributed conductance G, in S/m

        Returns
        -------        
        G : number, or array-like
            distributed conductance, in S/m

        '''
        f =  self.frequency.f
        return f*self.epsilon_second/log(self.b/self.a)

    def __str__(self):
        f=self.frequency
        try:
            output =  \
                'Coaxial Transmission Line.  %i-%i %s.  %i points'%\
                (f.f_scaled[0],f.f_scaled[-1],f.unit, f.npoints) + \
                '\nDint= %.2f mm, Dout= %.2f mm '% \
                (self.Dint*1e3, self.Dout*1e3) +\
                '\nCharacteristic Impedance=(%.1f,%.1fj)-(%.1f,%.1fj) Ohm'% \
                (real(self.Z0[0]), imag(self.Z0[0]), real(self.Z0[-1]), imag(self.Z0[-1])) +\
                '\nPort impedance Z0=(%.1f,%.1fj)-(%.1f,%.1fj) Ohm'% \
                (real(self.z0[0]), imag(self.z0[0]), real(self.z0[-1]), imag(self.z0[-1]))
        except(TypeError):
            output =  \
                'Coaxial Transmission Line.  %i-%i %s.  %i points'%\
                (f.f_scaled[0],f.f_scaled[-1],f.unit, f.npoints) + \
                '\nDint= %.2f mm, Dout= %.2f mm '% \
                (self.Dint[0]*1e3, self.Dout[0]*1e3) +\
                '\nCharacteristic Impedance=(%.1f,%.1fj)-(%.1f,%.1fj) Ohm'% \
                (real(self.Z0[0]), imag(self.Z0[0]), real(self.Z0[-1]), imag(self.Z0[-1])) +\
                '\nPort impedance Z0=(%.1f,%.1fj)-(%.1f,%.1fj) Ohm'% \
                (real(self.z0[0]), imag(self.z0[0]), real(self.z0[-1]), imag(self.z0[-1]))
        return output



from bokeh import models, plotting

from .utils import trace_color_cycle
from .. import network

plotting.output_notebook()

Y_LABEL_DICT = network.Y_LABEL_DICT  # type: dict
PRIMARY_PROPERTIES = network.PRIMARY_PROPERTIES  # type: dict
COMPONENT_FUNC_DICT = network.COMPONENT_FUNC_DICT  # type: dict

default_kwargs = {
    'primary_property': "s",
    'property_type': "db",
    'show': True,
    'fig': None
}


def plot_rectangular(ntwk, **kwargs):
    """
    :type ntwk: Network
    :return: plotting.figure.Figure
    """

    fig = kwargs.get("fig", None)
    show = kwargs.get("show", True)

    primary_property = kwargs.get("primary_property", "s")
    property_type = kwargs.get("property_type", "db")

    colors = trace_color_cycle()

    if type(fig) is not plotting.Figure:
        fig = plotting.figure(
            title=ntwk.name,
            height=350, width=800,
            x_axis_label="frequency ({:s})".format(ntwk.frequency.unit),
            y_axis_label=Y_LABEL_DICT[property_type],
            tools="resize, pan, wheel_zoom, box_zoom, save, reset",
            toolbar_location="above",
            toolbar_sticky=True
        )

    labels = []
    glpyhs = []

    for n in range(ntwk.nports):
        for m in range(ntwk.nports):
            x = ntwk.frequency.f_scaled
            y = getattr(ntwk, primary_property + "_" + property_type)[:, m, n]
            glpyhs.append(fig.line(x, y, line_color=next(colors)))
            labels.append("S{:d}{:d}".format(n + 1, m + 1))

    legend_items = []
    for label, glyph in zip(labels, glpyhs):
        legend_items.append((label, [glyph]))

    legend = models.Legend(items=legend_items, location=(0, -30))

    fig.add_layout(legend, 'right')

    if show:
        plotting.show(fig)

    return fig


def plot_polar():
    pass  # not native to bokeh, but I have seen some hacks to do this.  Smith chart may be tricky


def use_bokeh():
    for p in PRIMARY_PROPERTIES:
        for t in COMPONENT_FUNC_DICT.keys():
            attribute_name = "plot_{:s}_{:s}".format(p, t)

            def gen_plot_function(p, t):
                def plot_function(ntwk, **kwargs):
                    kwargs["primary_property"] = p
                    kwargs["property_type"] = t
                    plot_rectangular(ntwk, **kwargs)
                return plot_function

            setattr(network.Network, attribute_name, gen_plot_function(p, t))

use_bokeh()  # this function can be called again if we need to switch plotting engines for some reason




from six.moves import xrange # for Python3 compatibility

from .frequency import Frequency
from .mathFunctions import *
from .plotting import plot_complex_rectangular,plot_rectangular, smith
from .util import get_fid, get_extn, find_nearest_index,slice_domain

from scipy import  signal
import numpy as npy
from numpy import fft
import matplotlib.pyplot as plb

from IPython.display import Image
from IPython.core.pylabtools import print_figure

from abc import ABCMeta, abstractmethod
from copy import deepcopy
import re



##

class Parameter(object):
    '''
    a complex network parameter
    '''

    def __init__(self,  network):
        self._network = network

    def __len__(self):
        '''
        length of frequency axis
        '''
        return len(self.val)

    def __getattr__(self,name):
        return getattr(self.val,name)

    def __getitem__(self,key):
        return self.val[key]

    @property
    def val(self):
        raise NotImplementedError()

    @property
    def _xaxis(self):return 'frequency'

    ## projections
    @property
    def re(self): return Re(self)
    @property
    def im(self): return Im(self)
    @property
    def mag(self): return Mag(self)
    @property
    def deg(self): return Deg(self)
    @property
    def rad(self): return Rad(self)
    @property
    def db10(self): return Db10(self)
    @property
    def db20(self): return Db20(self)
    @property
    def db(self): return Db20(self)

    def plot(self, m=None, n=None, ax=None, show_legend=True,*args,
             **kwargs):

        # create index lists, if not provided by user
        if m is None:
            M = range(self._network.nports)
        else:
            M = [m]
        if n is None:
            N = range(self._network.nports)
        else:
            N = [n]

        if 'label'  not in kwargs.keys():
            gen_label = True
        else:
            gen_label = False


        #was_interactive = plb.isinteractive
        #if was_interactive:
        #    plb.interactive(False)
        lines = []
        for m in M:
            for n in N:
                # set the legend label for this trace to the networks
                # name if it exists, and they didn't pass a name key in
                # the kwargs
                if gen_label:
                    if self._network.name is None:
                        if plb.rcParams['text.usetex']:
                            label_string = '$%s_{%i%i}$'%\
                            (str(self).upper(),m+1,n+1)
                        else:
                            label_string = '%s%i%i'%\
                            (str(self).upper(),m+1,n+1)
                    else:
                        if plb.rcParams['text.usetex']:
                            label_string = str(self._network)+', $%s_{%i%i}$'%\
                            (str(self).upper(),m+1,n+1)
                        else:
                            label_string = self._network.name+', %s%i%i'%\
                            (str(self).upper(),m+1,n+1)
                    kwargs['label'] = label_string

                # plot the desired attribute vs frequency
                lines.append(plot_complex_rectangular(
                    z = self.val[:,m,n],
                    show_legend = show_legend, ax = ax,
                    *args, **kwargs))#[0]) ## fix
        #return lines ## fix
    def plot_smith(self, **kwargs):
        self.plot(**kwargs)
        smith()


    ## notebook display
    def _figure_data(self, format):
        fig, ax = plb.subplots()
        self.plot(ax=ax)
        data = print_figure(fig, format)
        plb.close(fig)
        return data

    def _repr_png_(self):
        return self._figure_data('png')

    @property
    def png(self):
        return Image(self._repr_png_(), embed=True)

class S(Parameter):
    '''
    S parameters

    This Parameter is special, because they are the internal storage format

    '''
    def __init__(self,  network, s):
        Parameter.__init__(self, network)
        s = fix_parameter_shape(s)
        self._val= npy.array(s,dtype=complex)

    def __getattr__(self,name):
        return getattr(self.val,name)

    def __str__(self): return 's'

    @property
    def val(self):
        return self._val


    def plot(self, *args, **kwargs):
        out = Parameter.plot(self,*args, **kwargs)
        smith()
        return out

    def plot_complex(self, *args, **kwargs):
        return Parameter.plot(self,*args, **kwargs)

class Z(Parameter):
    '''
    Impedance parameters
    '''
    def __str__(self): return 'z'
    @property
    def val(self):
        return s2z(self._network.s.val)

class Y(Parameter):
    '''
    Admittance Parameters
    '''
    def __str__(self): return 'y'
    @property
    def val(self):
        return s2y(self._network.s.val)

class T(Parameter):
    '''
    Wave Cascading Parameters

    Only exists for 2-ports
    '''
    def __str__(self): return 't'
    @property
    def val(self):
        return s2t(self._network.s.val)

class STime(Parameter):
    '''
    Scattering Parameters in Time Domain
    '''
    def __str__(self): return 's'
    @property
    def _xaxis(self):return 'time'
    @property
    def val(self):
        return s2time(self._network.s.val)
##

class Projection(object):
    '''
    a scalar projection of a parameter
    '''
    def __init__(self, param):
        self._param = param
        self._network = param._network

    def __getitem__(self,key):
        return self.val[key]

    def __getattr__(self,name):
        return getattr(self.val,name)

    @property
    def val(self):
        raise NotImplementedError()

    def plot(self,  m=None, n=None, ax=None, show_legend=True,*args,
             **kwargs):

        # create index lists, if not provided by user
        if m is None:
            M = range(self._network.nports)
        else:
            M = [m]
        if n is None:
            N = range(self._network.nports)
        else:
            N = [n]

        if 'label'  not in kwargs.keys():
            gen_label = True
        else:
            gen_label = False

        lines = [] #list of mpl lines
        for m in M:
            for n in N:
                # set the legend label for this trace to the networks
                # name if it exists, and they didnt pass a name key in
                # the kwargs
                if gen_label:
                    if self._network.name is None:
                        if plb.rcParams['text.usetex']:
                            label_string = '$%s_{%i%i}$'%\
                            (str(self._param).upper(),m+1,n+1)
                        else:
                            label_string = '%s%i%i'%\
                            (str(self._param).upper(),m+1,n+1)
                    else:
                        if plb.rcParams['text.usetex']:
                            label_string = self._network.name+', $%s_{%i%i}$'%\
                            (str(self._param).upper(),m+1,n+1)
                        else:
                            label_string = self._network.name+', %s%i%i'%\
                            (str(self._param).upper(),m+1,n+1)
                    kwargs['label'] = label_string

                # plot the desired attribute vs frequency
                if  self._param._xaxis=='time':
                    x_label = 'Time (ns)'
                    x = self._network.frequency.t_ns

                elif self._param._xaxis=='frequency':
                    x_label = 'Frequency (%s)'%self._network.frequency.unit
                    x = self._network.frequency.f_scaled


                lines.append(plot_rectangular(
                        x = x,
                        y = self.val[:,m,n],
                        x_label = x_label,
                        y_label = self.y_label,
                        show_legend = show_legend, ax = ax,
                        *args, **kwargs)[0])
        return lines

    def _figure_data(self, format):
        fig, ax = plb.subplots()
        self.plot(ax=ax)
        data = print_figure(fig, format)
        plb.close(fig)
        return data

    def _repr_png_(self):
        return self._figure_data('png')

    @property
    def png(self):
        return Image(self._repr_png_(), embed=True)

class Mag(Projection):
    y_label = 'Magnitude'
    unit=''
    def __str__(self):
        return ''
    def __repr__(self):
        return '{self._param}{self}'.format(self=self)

    @property
    def val(self):
        return abs(self._param.val)

class Db10(Projection):
    y_label = 'Magnitude (dB)'
    unit='dB'
    def __str__(self):
        return 'dB'
    def __repr__(self):
        return '{self._param}{self}'.format(self=self)

    @property
    def val(self):
        return complex_2_db10(self._param.val)

class Db20(Projection):
    y_label = 'Magnitude (dB)'
    unit = 'dB'
    def __str__(self):
        return 'dB'
    def __repr__(self):
        return '{self._param}{self}'.format(self=self)
    @property
    def val(self):
        return complex_2_db(self._param.val)

class Deg(Projection):
    y_label = 'Phase (deg)'
    unit = 'deg'
    def __str__(self):
        return 'deg'
    def __repr__(self):
        return '{self._param}{self}'.format(self=self)
    @property
    def val(self):
        return complex_2_degree(self._param.val)

class Rad(Projection):
    y_label = 'Phase (rad)'
    unit = 'rad'
    def __str__(self):
        return 'rad'
    def __repr__(self):
        return '{self._param}{self}'.format(self=self)
    @property
    def val(self):
        return complex_2_radian(self._param.val)

class Re(Projection):
    y_label = 'Real Part'
    unit = ''
    def __str__(self):
        return 'real'
    def __repr__(self):
        return '{self._param}{self}'.format(self=self)
    @property
    def val(self):
        return self._param.val.real

class Im(Projection):
    y_label = 'Imaginary Part'
    unit = ''
    def __str__(self):
        return 'imag'
    def __repr__(self):
        return '{self._param}{self}'.format(self=self)
    @property
    def val(self):
        return self._param.val.imag

##

class Network(object):
    def __init__(self, frequency=None, z0=50, name='', comments='',
                 *args,  **kw):
        '''
        '''
        if 's' in kw:
            self.s = kw['s']
        elif 'z' in kw:
            self.s = z2s(kw['z'],z0)
        elif 'y' in kw:
            self.s = y2s(kw['y'],z0)
        else:
            s=zeros(len(frequency))

        self.frequency = frequency
        self.z0 = z0
        self.name = name
        self.comments = comments

    @classmethod
    def from_ntwkv1( cls, network):
        return cls(frequency = network.frequency,
                   s = network.s,
                   z0 = network.z0,
                   name = network.name,
                   comments = network.comments,
                   )
    def __str__(self):
        f = self.frequency
        if self.name is None:
            name = ''
        else:
            name = self.name

        if len(npy.shape(self.z0)) == 0:
            z0 = str(self.z0)
        else:
            z0 = str(self.z0[0,:])

        output = '%i-Port Network: \'%s\',  %s, z0=%s' % (self.nports, name, str(f), z0)

        return output

    def __repr__(self):
        return self.__str__()

    def __call__(self, i,j):
        n = self.copy()
        n.s = n.s[:,i,j]
        return n

    def __len__(self):
        '''
        length of frequency axis
        '''
        return len(self.frequency)

    def __getitem__(self,key):
        '''
        Slices a Network object based on an index, or human readable string

        Parameters
        -----------
        key : str, or int
            if int; then it is interpreted as the index of the frequency
            if str, then should be like '50.1-75.5ghz', or just '50'.
            If the frequency unit is omitted then self.frequency.unit is
            used.

        Examples
        -----------
        >>> from skrf.data import ring_slot
        >>> a = ring_slot['80-90ghz']
        >>> a.plot_s_db()
        '''

        if isinstance(key, str):
            # they passed a string. try to read the string and convert
            # it into a  slice. then slice self on that
            re_numbers = re.compile('.*\d')
            re_hyphen = re.compile('\s*-\s*')
            re_letters = re.compile('[a-zA-Z]+')

            freq_unit = re.findall(re_letters,key)

            if len(freq_unit) == 0:
                freq_unit = self.frequency.unit
            else:
                freq_unit = freq_unit[0]

            key_nounit = re.sub(re_letters,'',key)
            edges  = re.split(re_hyphen,key_nounit)

            edges_freq = Frequency.from_f([float(k) for k in edges],
                                        unit = freq_unit)
            if len(edges_freq) ==2:
                slicer=slice_domain(self.frequency.f, edges_freq.f)
            elif len(edges_freq)==1:
                key = find_nearest_index(self.frequency.f, edges_freq.f[0])
                slicer = slice(key,key+1,1)
            else:
                raise ValueError()

            key = slicer

        try:

            output = self.copy()
            output.frequency.f = npy.array(output.frequency.f[key]).reshape(-1)
            output.z0 = output.z0[key,:]
            output.s = output.s[key,:,:]
            return output

        except(IndexError):
            raise IndexError('slicing frequency/index is incorrect')



    def copy(self):
        ntwk = Network(frequency =self.frequency.copy(),
                       s = self.s.val.copy(),
                       z0 = self.z0.copy(),
                       name = self.name,
                       comments = self.comments,
                       )

        return ntwk

    @property
    def s(self):
        '''
        Scattering Parameters
        '''
        return self._s

    @s.setter
    def s(self,s):
        self._s = S(self, s)

    @property
    def z(self):
        return Z(self)

    @z.setter
    def z(self,z):
        self.s = z2s(z,self.z0)

    @property
    def y(self):
        return Y(self)

    @y.setter
    def y(self,y):
        self.s = y2s(y,self.z0)

    @property
    def t(self):
        return T(self)

    @t.setter
    def t(self,t):
        raise NotImplementedError()

    @property
    def s_time(self):
        return STime(self)

    @s_time.setter
    def s_time(self):
        raise NotImplementedError()





    @property
    def nports(self):
        '''
        the number of ports the network has.

        Returns
        --------
        nports : number
                the number of ports the network has.

        '''
        try:
            return self.s.val.shape[1]
        except (AttributeError):
            return 0

    @property
    def z0(self):
        '''
        The port impedance
        '''
        return self._z0

    @z0.setter
    def z0(self,z0):
        self._z0 = fix_z0_shape(z0, len(self.frequency),nports=self.nports)

    @property
    def port_tuples(self):
        '''
        Returns a list of tuples, for each port index pair

        A convenience function for the common task fo iterating over
        all s-parameters index pairs

        This just calls:
        `[(y,x) for x in range(self.nports) for y in range(self.nports)]`
        '''
        return [(y,x) for x in range(self.nports) for y in range(self.nports)]

    def windowed(self, window=('kaiser',6),  normalize = True):
        '''
        Return a windowed version of s-matrix. Used in time-domain analysis.

        When using time domain through :attr:`s_time_db`,
        or similar properties, the spectrum is ussually windowed,
        before the IFFT is taken. This is done to
        compensate for the band-pass nature of a spectrum [1]_ .

        This function calls :func:`scipy.signal.get_window` which gives
        more details about the windowing.

        Parameters
        -----------
        window : string, float, or tuple
            The type of window to create. See :func:`scipy.signal.get_window`
            for details.
        normalize : bool
            Normalize the window to preserve power. ie
            sum(ntwk.s,axis=0) == sum(ntwk.windowed().s,axis=0)

        Examples
        -----------
        >>> ntwk = rf.Network('myfile.s2p')
        >>> ntwk_w = ntwk.windowed()
        >>> ntwk_w.plot_s_time_db()

        References
        -------------
        .. [1] Agilent Time Domain Analysis Using a Network Analyzer Application Note 1287-12

        '''

        windowed = self.copy()
        window = signal.get_window(window, len(self))
        window =window.reshape(-1,1,1) * npy.ones((len(self),
                                                   self.nports,
                                                   self.nports))
        windowed.s  = windowed.s[:] * window
        if normalize:
            # normalize the s-parameters to account for power lost in windowing
            windowed.s = windowed.s[:] * npy.sum(self.s.mag[:],axis=0)/\
                npy.sum(windowed.s.mag[:],axis=0)

        return windowed



def fix_z0_shape( z0, nfreqs, nports):
    '''
    Make a port impedance of correct shape for a given network's matrix

    This attempts to broadcast z0 to satisfy
        npy.shape(z0) == (nfreqs,nports)

    Parameters
    --------------
    z0 : number, array-like
        z0 can be:
        * a number (same at all ports and frequencies)
        * an array-like of length == number ports.
        * an array-like of length == number frequency points.
        * the correct shape ==(nfreqs,nports)

    nfreqs : int
        number of frequency points
    nportrs : int
        number of ports

    Returns
    ----------
    z0 : array of shape ==(nfreqs,nports)
        z0  with the right shape for a nport Network

    Examples
    ----------
    For a two-port network with 201 frequency points, possible uses may
    be

    >>> z0 = rf.fix_z0_shape(50 , 201,2)
    >>> z0 = rf.fix_z0_shape([50,25] , 201,2)
    >>> z0 = rf.fix_z0_shape(range(201) , 201,2)


    '''



    if npy.shape(z0) == (nfreqs, nports):
        # z0 is of correct shape. super duper.return it quick.
        return z0.copy()

    elif npy.isscalar(z0):
        # z0 is a single number
        return npy.array(nfreqs*[nports * [z0]])

    elif len(z0)  == nports:
        # assume z0 is a list of impedances for each port,
        # but constant with frequency
        return npy.array(nfreqs*[z0])

    elif len(z0) == nfreqs:
        # assume z0 is a list of impedances for each frequency,
        # but constant with respect to ports
        return npy.array(nports * [z0]).T

    else:
        raise IndexError('z0 is not acceptable shape')

def fix_parameter_shape(s):
    s_shape= npy.shape(s)
    if len(s_shape) <3:
        if len(s_shape) == 2:
            # reshape to kx1x1, this simplifies indexing in function
            s = npy.reshape(s,(-1,s_shape[0],s_shape[0]))
        else:
            s = npy.reshape(s,(-1,1,1))
    return s

## network parameter conversion

def s2z(s,z0=50):
    '''
    Convert scattering parameters [1]_ to impedance parameters [2]_


    .. math::
        z = \\sqrt {z_0} \\cdot (I + s) (I - s)^{-1} \\cdot \\sqrt{z_0}

    Parameters
    ------------
    s : complex array-like
        scattering parameters
    z0 : complex array-like or number
        port impedances.

    Returns
    ---------
    z : complex array-like
        impedance parameters



    References
    ----------
    .. [1] http://en.wikipedia.org/wiki/S-parameters
    .. [2] http://en.wikipedia.org/wiki/impedance_parameters

    '''
    s = s.copy() # to prevent the original array from being altered
    s = fix_parameter_shape(s)
    nfreqs, nports, nports = s.shape
    z0 = fix_z0_shape(z0, nfreqs, nports)

    z = npy.zeros(s.shape, dtype='complex')
    I = npy.mat(npy.identity(s.shape[1]))

    s[s==1.] = 1. + 1e-12 # solve numerical singularity
    s[s==-1.] = -1. + 1e-12 # solve numerical singularity
    for fidx in xrange(s.shape[0]):
        sqrtz0 = npy.mat(npy.sqrt(npy.diagflat(z0[fidx])))
        z[fidx] = sqrtz0 * (I-s[fidx])**-1 * (I+s[fidx]) * sqrtz0
    return z

def s2y(s,z0=50):
    '''
    convert scattering parameters [#]_ to admittance parameters [#]_


    .. math::
        y = \\sqrt {y_0} \\cdot (I - s)(I + s)^{-1} \\cdot \\sqrt{y_0}

    Parameters
    ------------
    s : complex array-like
        scattering parameters
    z0 : complex array-like or number
        port impedances

    Returns
    ---------
    y : complex array-like
        admittance parameters

    See Also
    ----------
    s2z
    s2y
    s2t
    z2s
    z2y
    z2t
    y2s
    y2z
    y2z
    t2s
    t2z
    t2y
    Network.s
    Network.y
    Network.z
    Network.t

    References
    ----------
    .. [#] http://en.wikipedia.org/wiki/S-parameters
    .. [#] http://en.wikipedia.org/wiki/Admittance_parameters
    '''
    s = s.copy() # to prevent the original array from being altered
    s = fix_parameter_shape(s)
    nfreqs, nports, nports = s.shape
    z0 = fix_z0_shape(z0, nfreqs, nports)
    y = npy.zeros(s.shape, dtype='complex')
    I = npy.mat(npy.identity(s.shape[1]))
    s[s==-1.] = -1. + 1e-12 # solve numerical singularity
    s[s==1.] = 1. + 1e-12 # solve numerical singularity
    for fidx in xrange(s.shape[0]):
        sqrty0 = npy.mat(npy.sqrt(npy.diagflat(1.0/z0[fidx])))
        y[fidx] = sqrty0*(I-s[fidx])*(I+s[fidx])**-1*sqrty0
    return y

def s2t(s):
    '''
    Converts scattering parameters [#]_ to scattering transfer parameters [#]_ .

    transfer parameters are also referred to as
    'wave cascading matrix', this function only operates on 2-port
    networks.

    Parameters
    -----------
    s : :class:`numpy.ndarray` (shape fx2x2)
        scattering parameter matrix

    Returns
    -------
    t : numpy.ndarray
        scattering transfer parameters (aka wave cascading matrix)

    See Also
    ---------
    inv : calculates inverse s-parameters

    s2z
    s2y
    s2t
    z2s
    z2y
    z2t
    y2s
    y2z
    y2z
    t2s
    t2z
    t2y
    Network.s
    Network.y
    Network.z
    Network.t

    References
    -----------
    .. [#] http://en.wikipedia.org/wiki/S-parameters
    .. [#] http://en.wikipedia.org/wiki/Scattering_transfer_parameters#Scattering_transfer_parameters
    '''
    #TODO: check rank(s) ==2
    s = s.copy() # to prevent the original array from being altered
    s = fix_parameter_shape(s)
    t = npy.array([
        [-1*(s[:,0,0]*s[:,1,1]- s[:,1,0]*s[:,0,1])/s[:,1,0],
            -s[:,1,1]/s[:,1,0]],
        [s[:,0,0]/s[:,1,0],
            1/s[:,1,0] ]
        ]).transpose()
    return t

def s2time(s,window =('kaiser',6),  normalize = True):
    '''
    '''
    s = s.copy() # to prevent the original array from being altered
    s = fix_parameter_shape(s)
    nfreqs, nports, nports = s.shape


    window = signal.get_window(window,nfreqs)
    window =window.reshape(-1,1,1) * npy.ones(s.shape)
    windowed = s * window
    if normalize:
        # normalize the s-parameters to account for power lost in windowing
        norm_factor = npy.sum(abs(s),axis=0)/\
                      npy.sum(abs(windowed),axis=0)
        windowed = windowed*norm_factor

    time = fft.ifftshift(fft.ifft(windowed, axis=0), axes=0)
    return time


def z2s(z, z0=50):
    '''
    convert impedance parameters [1]_ to scattering parameters [2]_

    .. math::
        s = (\\sqrt{y_0} \\cdot z \\cdot \\sqrt{y_0} - I)(\\sqrt{y_0} \\cdot z \\cdot\\sqrt{y_0} + I)^{-1}

    Parameters
    ------------
    z : complex array-like
        impedance parameters
    z0 : complex array-like or number
        port impedances

    Returns
    ---------
    s : complex array-like
        scattering parameters



    References
    ----------
    .. [1] http://en.wikipedia.org/wiki/impedance_parameters
    .. [2] http://en.wikipedia.org/wiki/S-parameters
    '''
    z = z.copy() # to prevent the original array from being altered
    z = fix_parameter_shape(z)
    nfreqs, nports, nports = z.shape
    z0 = fix_z0_shape(z0, nfreqs, nports)
    s = npy.zeros(z.shape, dtype='complex')
    I = npy.mat(npy.identity(z.shape[1]))
    for fidx in xrange(z.shape[0]):
        sqrty0 = npy.mat(npy.sqrt(npy.diagflat(1.0/z0[fidx])))
        s[fidx] = (sqrty0*z[fidx]*sqrty0 - I) * (sqrty0*z[fidx]*sqrty0 + I)**-1
    return s

def z2y(z):
    '''
    convert impedance parameters [#]_ to admittance parameters [#]_


    .. math::
        y = z^{-1}

    Parameters
    ------------
    z : complex array-like
        impedance parameters

    Returns
    ---------
    y : complex array-like
        admittance parameters

    See Also
    ----------
    s2z
    s2y
    s2t
    z2s
    z2y
    z2t
    y2s
    y2z
    y2z
    t2s
    t2z
    t2y
    Network.s
    Network.y
    Network.z
    Network.t

    References
    ----------
    .. [#] http://en.wikipedia.org/wiki/impedance_parameters
    .. [#] http://en.wikipedia.org/wiki/Admittance_parameters
    '''
    z = z.copy() # to prevent the original array from being altered
    z = fix_parameter_shape(z)
    return npy.array([npy.mat(z[f,:,:])**-1 for f in xrange(z.shape[0])])

def z2t(z):
    '''
    Not Implemented yet

    convert impedance parameters [#]_ to scattering transfer parameters [#]_


    Parameters
    ------------
    z : complex array-like or number
        impedance parameters

    Returns
    ---------
    s : complex array-like or number
        scattering parameters

    See Also
    ----------
    s2z
    s2y
    s2t
    z2s
    z2y
    z2t
    y2s
    y2z
    y2z
    t2s
    t2z
    t2y
    Network.s
    Network.y
    Network.z
    Network.t


    References
    ----------
    .. [#] http://en.wikipedia.org/wiki/impedance_parameters
    .. [#] http://en.wikipedia.org/wiki/Scattering_transfer_parameters#Scattering_transfer_parameters
    '''
    raise (NotImplementedError)

def y2s(y, z0=50):
    '''
    convert admittance parameters [#]_ to scattering parameters [#]_


    .. math::
        s = (I - \\sqrt{z_0} \\cdot y \\cdot \\sqrt{z_0})(I + \\sqrt{z_0} \\cdot y \\cdot \\sqrt{z_0})^{-1}

    Parameters
    ------------
    y : complex array-like
        admittance parameters

    z0 : complex array-like or number
        port impedances

    Returns
    ---------
    s : complex array-like or number
        scattering parameters

    See Also
    ----------
    s2z
    s2y
    s2t
    z2s
    z2y
    z2t
    y2s
    y2z
    y2z
    t2s
    t2z
    t2y
    Network.s
    Network.y
    Network.z
    Network.t


    References
    ----------
    .. [#] http://en.wikipedia.org/wiki/Admittance_parameters
    .. [#] http://en.wikipedia.org/wiki/S-parameters
    '''
    y = y.copy() # to prevent the original array from being altered
    y = fix_parameter_shape(y)
    nfreqs, nports, nports = y.shape
    z0 = fix_z0_shape(z0, nfreqs, nports)
    s = npy.zeros(y.shape, dtype='complex')
    I = npy.mat(npy.identity(s.shape[1]))
    for fidx in xrange(s.shape[0]):
        sqrtz0 = npy.mat(npy.sqrt(npy.diagflat(z0[fidx])))
        s[fidx] = (I - sqrtz0*y[fidx]*sqrtz0) * (I + sqrtz0*y[fidx]*sqrtz0)**-1
    return s

def y2z(y):
    '''
    convert admittance parameters [#]_ to impedance parameters [#]_


    .. math::
        z = y^{-1}

    Parameters
    ------------
    y : complex array-like
        admittance parameters

    Returns
    ---------
    z : complex array-like
        impedance parameters

    See Also
    ----------
    s2z
    s2y
    s2t
    z2s
    z2y
    z2t
    y2s
    y2z
    y2z
    t2s
    t2z
    t2y
    Network.s
    Network.y
    Network.z
    Network.t

    References
    ----------
    .. [#] http://en.wikipedia.org/wiki/Admittance_parameters
    .. [#] http://en.wikipedia.org/wiki/impedance_parameters
    '''
    y = y.copy() # to prevent the original array from being altered
    y = fix_parameter_shape(y)
    return npy.array([npy.mat(y[f,:,:])**-1 for f in xrange(y.shape[0])])

def y2t(y):
    '''
    Not Implemented Yet

    convert admittance parameters [#]_ to scattering-transfer parameters [#]_


    Parameters
    ------------
    y : complex array-like or number
        impedance parameters

    Returns
    ---------
    t : complex array-like or number
        scattering parameters

    See Also
    ----------
    s2z
    s2y
    s2t
    z2s
    z2y
    z2t
    y2s
    y2z
    y2z
    t2s
    t2z
    t2y
    Network.s
    Network.y
    Network.z
    Network.t

    References
    ----------
    .. [#] http://en.wikipedia.org/wiki/Admittance_parameters
    .. [#] http://en.wikipedia.org/wiki/Scattering_transfer_parameters#Scattering_transfer_parameters
    '''
    raise (NotImplementedError)

def t2s(t):
    '''
    converts scattering transfer parameters [#]_ to scattering parameters [#]_

    transfer parameters are also refered to as
    'wave cascading matrix', this function only operates on 2-port
    networks. this function only operates on 2-port scattering
    parameters.

    Parameters
    -----------
    t : :class:`numpy.ndarray` (shape fx2x2)
            scattering transfer parameters

    Returns
    -------
    s : :class:`numpy.ndarray`
            scattering parameter matrix.

    See Also
    ---------
    inv : calculates inverse s-parameters
    s2z
    s2y
    s2t
    z2s
    z2y
    z2t
    y2s
    y2z
    y2z
    t2s
    t2z
    t2y
    Network.s
    Network.y
    Network.z
    Network.t

    References
    -----------
    .. [#] http://en.wikipedia.org/wiki/Scattering_transfer_parameters#Scattering_transfer_parameters
    .. [#] http://en.wikipedia.org/wiki/S-parameters
    '''
    #TODO: check rank(s) ==2
    t = t.copy() # to prevent the original array from being altered
    t = fix_parameter_shape(t)
    s = npy.array([
        [t[:,0,1]/t[:,1,1],
             1/t[:,1,1]],
        [(t[:,0,0]*t[:,1,1]- t[:,1,0]*t[:,0,1])/t[:,1,1],
            -1*t[:,1,0]/t[:,1,1] ]
        ]).transpose()
    return s

def t2z(t):
    '''
    Not Implemented  Yet

    Convert scattering transfer parameters [#]_ to impedance parameters [#]_



    Parameters
    ------------
    t : complex array-like or number
        impedance parameters

    Returns
    ---------
    z : complex array-like or number
        scattering parameters

    See Also
    ----------
    s2z
    s2y
    s2t
    z2s
    z2y
    z2t
    y2s
    y2z
    y2z
    t2s
    t2z
    t2y
    Network.s
    Network.y
    Network.z
    Network.t

    References
    ----------
    .. [#] http://en.wikipedia.org/wiki/Scattering_transfer_parameters#Scattering_transfer_parameters
    .. [#] http://en.wikipedia.org/wiki/impedance_parameters
    '''
    raise (NotImplementedError)

def t2y(t):
    '''
    Not Implemented Yet

    Convert scattering transfer parameters to admittance parameters [#]_




    Parameters
    ------------
    t : complex array-like or number
        t-parameters

    Returns
    ---------
    y : complex array-like or number
        admittance parameters

    See Also
    ----------
    s2z
    s2y
    s2t
    z2s
    z2y
    z2t
    y2s
    y2z
    y2z
    t2s
    t2z
    t2y
    Network.s
    Network.y
    Network.z
    Network.t

    References
    ----------
    .. [#] http://en.wikipedia.org/wiki/Scattering_transfer_parameters#Scattering_transfer_parameters

    '''
    raise (NotImplementedError)
